"""
Transformations and augmentations to be applied to echogram transects.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import collections
import os
import random
import warnings

import numpy as np
import scipy.interpolate
import scipy.ndimage.filters
import skimage.transform

from echofilter.nn.modules.utils import _pair

_fields_2d = (
    "Sv",
    "signals",
    "mask",
    "mask_turbulence",
    "mask_top",
    "mask_bottom",
    "mask_bot",
    "mask_turbulence-original",
    "mask_top-original",
    "mask_bottom-original",
    "mask_bot-original",
    "mask_surface",
    "mask_surf",
    "mask_patches",
    "mask_patches-original",
    "mask_patches-ntob",
)
_fields_1d_timelike = (
    "timestamps",
    "turbulence",
    "top",
    "bottom",
    "turbulence-original",
    "top-original",
    "bottom-original",
    "surface",
    "surf",
    "d_turbulence",
    "d_top",
    "d_bottom",
    "d_bot",
    "r_turbulence",
    "r_top",
    "r_bottom",
    "r_bot",
    "d_turbulence-original",
    "d_top-original",
    "d_bottom-original",
    "d_bot-original",
    "r_turbulence-original",
    "r_top-original",
    "r_bottom-original",
    "r_bot-original",
    "d_surface",
    "d_surf",
    "r_surface",
    "r_surf",
    "is_surrogate_surface",
    "is_bad_labels",
    "is_passive",
    "is_removed",
)
_fields_1d_depthlike = ("depths",)
_fields_0d = ("is_upward_facing",)
_fields_needing_linear = (
    "timestamps",
    "depths",
    "is_surrogate_surface",
    "is_bad_labels",
    "is_passive",
    "is_removed",
)


class Rescale(object):
    """
    Rescale the image(s) in a sample to a given size.

    Parameters
    ----------
    output_size : tuple or int
        Desired output size. If tuple, output is matched to output_size. If
        int, output is square.
    order : int or None, optional
        Order of the interpolation, for both image and vector elements.
        For images-like components, the interpolation is 2d.
        The following values are supported:

        - 0: Nearest-neighbor
        - 1: Linear (default)
        - 2: Quadratic
        - 3: Cubic

        If ``None``, the order is randomly selected as either ``0`` or ``1``.
    """

    order2kind = {
        0: "nearest",
        1: "linear",
        2: "quadratic",
        3: "cubic",
    }

    def __init__(self, output_size, order=1):
        self.output_size = _pair(output_size)
        self.order = order

    def __call__(self, sample):

        order = self.order
        if order is None:
            order = np.random.randint(1)

        kind = self.order2kind[order]

        # 2D arrays (image-like)
        for key in _fields_2d:
            if key not in sample:
                continue
            if sample[key].shape == self.output_size:
                continue
            _dtype = sample[key].dtype
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", "Bi-quadratic interpolation behavior has changed"
                )
                sample[key] = skimage.transform.resize(
                    np.asarray(sample[key]).astype(np.float),
                    self.output_size,
                    order=order,
                    clip=False,
                    preserve_range=False,
                )
            sample[key] = sample[key].astype(_dtype)

        # 1D arrays (column-like)
        for key in _fields_1d_timelike:
            if key not in sample:
                continue
            if sample[key].shape == self.output_size[:1]:
                continue
            _kind = "linear" if key == "timestamps" else kind
            if order > 1 and (
                key in _fields_needing_linear
                or ("bot" in key and sample["is_upward_facing"])
            ):
                _kind = "linear"
            _dtype = sample[key].dtype
            sample[key] = scipy.interpolate.interp1d(
                np.arange(len(sample[key])),
                sample[key],
                kind=_kind,
            )(np.linspace(0, len(sample[key]) - 1, self.output_size[0]))
            sample[key] = sample[key].astype(_dtype)

        # 1D arrays (row-like)
        for key in _fields_1d_depthlike:
            if key not in sample:
                continue
            if sample[key].shape == self.output_size[1:]:
                continue
            _kind = "linear" if key == "depths" else kind
            _dtype = sample[key].dtype
            sample[key] = scipy.interpolate.interp1d(
                np.arange(len(sample[key])),
                sample[key],
                kind=_kind,
            )(np.linspace(0, len(sample[key]) - 1, self.output_size[1]))
            sample[key] = sample[key].astype(_dtype)

        return sample


class RandomGridSampling(Rescale):
    """
    Resample data onto a new grid, which is randomly resampled.

    Parameters
    ----------
    output_size : tuple or int
        Desired output size. If tuple, output is matched to output_size. If
        int, output is square.
    p : float, optional
        Probability of performing the RandomGrid operation. Default is ``0.5``.
    order : int or None, optional
        Order of the interpolation, for both image and vector elements.
        For images-like components, the interpolation is 2d.
        The following values are supported:

        - 0: Nearest-neighbor
        - 1: Linear (default)
        - 2: Quadratic
        - 3: Cubic

        If ``None``, the order is randomly selected from the set
        ``{0, 1, 3}``.
    """

    def __init__(self, *args, p=0.5, **kwargs):
        super(RandomGridSampling, self).__init__(*args, **kwargs)
        self.p = p

    def __call__(self, sample):

        if random.random() > self.p:
            # Nothing to do
            return sample

        order = self.order
        if order is None:
            # Randomly sample 0, 1, or 3
            order = np.random.randint(3)
            if order == 2:
                order += 1

        kind = self.order2kind[order]

        # Randomly re-sample x and y
        nx = len(sample["timestamps"])
        ny = len(sample["depths"])
        x_out = np.sort(np.random.uniform(0, nx - 1, size=self.output_size[0]))
        y_out = np.sort(np.random.uniform(0, ny - 1, size=self.output_size[1]))

        # 2D arrays (image-like)
        for key in _fields_2d:
            if key not in sample:
                continue
            _dtype = sample[key].dtype
            sample[key] = np.asarray(sample[key])
            if order == 0:
                if sample[key].shape != (nx, ny):
                    raise ValueError(
                        'Expected sample["{}"] to be shaped {}'.format(key, (nx, ny))
                    )
                sample[key] = sample[key][np.round(x_out).astype(int)][
                    :, np.round(y_out).astype(int)
                ]
            else:
                sample[key] = scipy.interpolate.RectBivariateSpline(
                    np.linspace(0, nx - 1, sample[key].shape[0]),
                    np.linspace(0, ny - 1, sample[key].shape[1]),
                    sample[key].astype(np.float),
                    kx=order,
                    ky=order,
                )(x_out, y_out)
            sample[key] = sample[key].astype(_dtype)

        # 1D arrays (column-like)
        for key in _fields_1d_timelike:
            if key not in sample:
                continue
            _kind = "linear" if key == "timestamps" else kind
            if order > 1 and (
                key in _fields_needing_linear
                or ("bot" in key and sample["is_upward_facing"])
            ):
                _kind = "linear"
            _dtype = sample[key].dtype
            sample[key] = scipy.interpolate.interp1d(
                np.linspace(0, nx - 1, len(sample[key])),
                sample[key],
                kind=_kind,
            )(x_out)
            sample[key] = sample[key].astype(_dtype)

        # 1D arrays (row-like)
        for key in _fields_1d_depthlike:
            if key not in sample:
                continue
            _kind = "linear" if key == "depths" else kind
            _dtype = sample[key].dtype
            sample[key] = scipy.interpolate.interp1d(
                np.linspace(0, ny - 1, len(sample[key])),
                sample[key],
                kind=_kind,
            )(y_out)
            sample[key] = sample[key].astype(_dtype)

        return sample


class RandomElasticGrid(Rescale):
    """
    Resample data onto a new grid, elastically deformed from the original grid.

    Parameters
    ----------
    output_size : tuple or int or None
        Desired output size. If tuple, output is matched to output_size. If
        int, output is square. If ``None``, the size remains unchanged from the
        input.
    p : float, optional
        Probability of performing the RandomGrid operation. Default is ``0.5``.
    sigma : float, optional
        Gaussian filter kernel size. Default is ``8.0``.
    alpha : float, optional
        Maximum size of image distortions, relative to the length of the side
        of the image. Default is ``0.05``.
    order : int or None, optional
        Order of the interpolation, for both image and vector elements.
        For images-like components, the interpolation is 2d.
        The following values are supported:

        - 0: Nearest-neighbor
        - 1: Linear (default)
        - 2: Quadratic
        - 3: Cubic

        If ``None``, the order is randomly selected from the set
        ``{1, 2, 3}``.
    """

    def __init__(self, output_size, p=0.5, sigma=8.0, alpha=0.05, order=1):
        self.output_size = None if output_size is None else _pair(output_size)
        self.order = order
        self.p = p
        self.sigma = _pair(sigma)
        self.alpha = _pair(alpha)

    def __call__(self, sample):

        if random.random() > self.p:
            # Nothing to do
            return sample

        order = self.order
        if order is None:
            # Randomly sample 1, 2 or 3
            order = random.randint(1, 3)

        kind = self.order2kind[order]

        # Randomly re-sample x and y
        nx = len(sample["timestamps"])
        ny = len(sample["depths"])
        output_size = self.output_size
        if output_size is None:
            output_size = (nx, ny)
        if nx < 2 or ny < 2:
            raise ValueError("Input image shape ({}, {}) is too small".format(nx, ny))
        if output_size[0] < 2 or output_size[1] < 2:
            raise ValueError("Output image shape {} is too small".format(output_size))

        x_out = np.linspace(0, nx - 1, output_size[0])
        x_intv = x_out[1] - x_out[0]
        dx = x_intv * 2 * (np.random.rand(*x_out.shape) - 0.5)
        dx = scipy.ndimage.filters.gaussian_filter1d(dx, self.sigma[0], cval=0)
        dx *= self.alpha[0] * nx
        x_out += dx

        y_out = np.linspace(0, ny - 1, output_size[1])
        y_intv = y_out[1] - y_out[0]
        dy = y_intv * 2 * (np.random.rand(*y_out.shape) - 0.5)
        dy = scipy.ndimage.filters.gaussian_filter1d(dy, self.sigma[1], cval=0)
        dy *= self.alpha[1] * ny
        y_out += dy

        x_out = x_out.clip(0, nx - 1)
        y_out = y_out.clip(0, ny - 1)
        x_out = np.sort(x_out)
        y_out = np.sort(y_out)

        # 2D arrays (image-like)
        for key in _fields_2d:
            if key not in sample:
                continue
            _dtype = sample[key].dtype
            sample[key] = np.asarray(sample[key])
            if order == 0:
                if sample[key].shape != (nx, ny):
                    raise ValueError(
                        'Expected sample["{}"] to be shaped {}'.format(key, (nx, ny))
                    )
                sample[key] = sample[key][np.round(x_out).astype(int)][
                    :, np.round(y_out).astype(int)
                ]
            else:
                sample[key] = scipy.interpolate.RectBivariateSpline(
                    np.linspace(0, nx - 1, sample[key].shape[0]),
                    np.linspace(0, ny - 1, sample[key].shape[1]),
                    sample[key].astype(np.float),
                    kx=order,
                    ky=order,
                )(x_out, y_out)
            sample[key] = sample[key].astype(_dtype)

        # 1D arrays (column-like)
        for key in _fields_1d_timelike:
            if key not in sample:
                continue
            _kind = "linear" if key == "timestamps" else kind
            if key in {"is_passive", "is_removed"} and order > 1:
                _kind = "linear"
            _dtype = sample[key].dtype
            sample[key] = scipy.interpolate.interp1d(
                np.linspace(0, nx - 1, len(sample[key])),
                sample[key],
                kind=_kind,
            )(x_out)
            sample[key] = sample[key].astype(_dtype)

        # 1D arrays (row-like)
        for key in _fields_1d_depthlike:
            if key not in sample:
                continue
            _kind = "linear" if key == "depths" else kind
            _dtype = sample[key].dtype
            sample[key] = scipy.interpolate.interp1d(
                np.linspace(0, ny - 1, len(sample[key])),
                sample[key],
                kind=_kind,
            )(y_out)
            sample[key] = sample[key].astype(_dtype)

        return sample


class Normalize(object):
    """
    Normalize offset and scaling of image (mean and standard deviation).

    Note that changes are made inplace.

    Parameters
    ----------
    center : {"mean", "median", "pc10"} or float
        If a float, a pre-computed centroid measure of the distribution of
        samples, such as the pixel mean. If a string, a method to use to
        determine the center value.
    deviation : {"stdev", "mad", "iqr", "idr", "i7r"} or float
        If a float, a pre-computed deviation measure of the distribution of
        samples. If a string, a method to use to determine the deviation.
    robust2stdev : bool, optional
        Whether to convert robust measures to estimates of the standard
        deviation. Default is ``True``.
    """

    def __init__(self, center, deviation, robust2stdev=True):
        self.center = center
        self.deviation = deviation
        self.robust2stdev = robust2stdev

    def __call__(self, sample):

        if not isinstance(self.center, str):
            center = self.center
        elif self.center.lower() == "mean":
            center = np.nanmean(sample["signals"])
        elif self.center.lower() == "median":
            center = np.nanmedian(sample["signals"])
        elif self.center.lower() == "pc10":
            center = np.nanpercentile(sample["signals"], 10)
        else:
            raise ValueError("Unrecognised center method: {}".format(self.center))

        if not isinstance(self.deviation, str):
            deviation = self.deviation
        elif self.deviation.lower() in {"std", "stdev"}:
            deviation = np.nanstd(sample["signals"])
        elif self.deviation.lower() == "mad":
            deviation = np.nanmedian(
                np.abs(sample["signals"] - np.nanmedian(sample["signals"]))
            )
            if self.robust2stdev:
                deviation *= 1.4826
        elif self.deviation.lower() == "iqr":
            deviation = np.diff(np.nanpercentile(sample["signals"], [25, 75])).item()
            if self.robust2stdev:
                deviation /= 1.35
        elif self.deviation.lower() == "idr":
            deviation = np.diff(np.nanpercentile(sample["signals"], [10, 90])).item()
            if self.robust2stdev:
                deviation /= 2.56
        elif self.deviation.lower() == "i7r":
            deviation = np.diff(np.nanpercentile(sample["signals"], [7, 93])).item()
            if self.robust2stdev:
                deviation /= 3.0
        else:
            raise ValueError("Unrecognised deviation method: {}".format(self.deviation))

        sample["signals"] -= center
        sample["signals"] /= deviation

        return sample


class ReplaceNan(object):
    """
    Replace NaNs with a finite float value.

    Parameters
    ----------
    nan_val : float, optional
        Value to replace NaNs with. Default is ``0.0``.
    """

    def __init__(self, nan_val=0.0):
        self.nan_val = nan_val

    def __call__(self, sample):

        # Can't use np.nan_to_num to assign nan to a specific value if
        # numpy version <= 1.17.
        sample["signals"][np.isnan(sample["signals"])] = self.nan_val

        return sample


class RandomReflection(object):
    """
    Randomly reflect a sample.

    Parameters
    ----------
    axis : int, optional
        Axis to reflect. Default is 0.
    p : float, optional
        Probability of reflection. Default is 0.5.
    """

    def __init__(self, axis=0, p=0.5):
        self.axis = axis
        self.p = p

    def __call__(self, sample):

        if random.random() > self.p:
            # Nothing to do
            return sample

        # Reflect data
        for key in _fields_2d + _fields_1d_timelike:
            if key in sample:
                sample[key] = np.flip(sample[key], self.axis).copy()

        return sample


class RandomCropWidth(object):
    """
    Randomly crop a sample in the width dimension.

    Parameters
    ----------
    max_crop_fraction : float
        Maximum amount of material to crop away, as a fraction of the total
        width. The ``crop_fraction`` will be sampled uniformly from the range
        ``[0, max_crop_fraction]``. The crop is always centred.
    """

    def __init__(self, max_crop_fraction):
        self.max_crop_fraction = max_crop_fraction

    def __call__(self, sample):

        width = sample["signals"].shape[0]

        crop_fraction = random.uniform(0.0, self.max_crop_fraction)
        crop_amount = crop_fraction * width

        lft = int(crop_amount / 2)
        rgt = lft + width - int(crop_amount)

        # Crop data
        for key in _fields_2d + _fields_1d_timelike:
            if key in sample:
                sample[key] = sample[key][lft:rgt]

        return sample


def optimal_crop_depth(transect):
    """
    Crop a sample depthwise to surround the water column.

    The crop is to contain only the space between highest surface
    and deepest seafloor.

    Parameters
    ----------
    transect : dict
        Transect dictionary.
    """
    d0 = np.min(transect["depths"])

    depth_intv = abs(transect["depths"][1] - transect["depths"][0])
    shallowest_depth = None
    if transect["is_upward_facing"]:
        for key in ("d_surface", "surface", "d_surf", "surf"):
            if key not in transect:
                continue
            surf_options = transect[key][transect[key] > d0]
            if len(surf_options) == 0:
                continue
            d = np.min(surf_options)
            if shallowest_depth is None:
                shallowest_depth = d
            else:
                shallowest_depth = min(d, shallowest_depth)
    if shallowest_depth is None:
        shallowest_depth = d0
    shallowest_depth -= 5 * depth_intv

    deepest_depth = None
    if not transect["is_upward_facing"]:
        for key in (
            "d_bottom-original",
            "d_bot-original",
            "d_bottom",
            "d_bot",
            "bottom-original",
            "bottom",
        ):
            if key not in transect:
                continue
            d = np.max(transect[key])
            if deepest_depth is None:
                deepest_depth = d
            else:
                deepest_depth = max(d, deepest_depth)
    if deepest_depth is None:
        deepest_depth = np.max(transect["depths"])
    deepest_depth += 5 * depth_intv

    if shallowest_depth >= deepest_depth:
        return transect

    depth_mask = (shallowest_depth <= transect["depths"]) & (
        transect["depths"] <= deepest_depth
    )

    for key in _fields_1d_depthlike:
        if key in transect:
            transect[key] = transect[key][depth_mask]

    for key in _fields_2d:
        if key in transect:
            transect[key] = transect[key][:, depth_mask]

    return transect


class OptimalCropDepth(object):
    """
    A transform which crops a sample depthwise to focus on the water column.

    The output contains only the space between highest surface and deepest
    seafloor line measurements.
    """

    def __call__(self, sample):
        return optimal_crop_depth(sample)


class RandomCropDepth(object):
    """
    Randomly crop a sample depthwise.

    Parameters
    ----------
    p_crop_is_none : float, optional
        Probability of not doing any crop.
        Default is ``0.1``.
    p_crop_is_optimal : float, optional
        Probability of doing an "optimal" crop, running ``optimal_crop_depth``.
        Default is ``0.1``.
    p_crop_is_close : float, optional
        Probability of doing crop which is zoomed in and close to the "optimal"
        crop, running ``optimal_crop_depth``. Default is ``0.4``.
        If neither no crop, optimal, nor close-to-optimal crop is selected,
        the crop is randomly sized over the full extent of the range of depths.
    p_nearfield_side_crop : float, optional
        Probability that the nearfield side is cropped.
        Default is ``0.5``.
    fraction_close : float, optional
        Fraction by which crop is increased/decreased in either direction when
        doing a close to optimal crop.
        Default is ``0.25``.
    """

    def __init__(
        self,
        p_crop_is_none=0.1,
        p_crop_is_optimal=0.1,
        p_crop_is_close=0.4,
        p_nearfield_side_crop=0.5,
        fraction_close=0.25,
    ):
        self.p_crop_is_none = p_crop_is_none
        self.p_crop_is_optimal = p_crop_is_optimal
        self.p_crop_is_close = p_crop_is_close
        self.p_nearfield_side_crop = p_nearfield_side_crop
        self.fraction_close = fraction_close

    def __call__(self, sample):

        p = random.random()

        # Possibility of doing no crop at all
        p -= self.p_crop_is_none
        if p < 0:
            return sample

        # Possibility of doing an optimal crop
        p -= self.p_crop_is_optimal
        if p < 0:
            return optimal_crop_depth(sample)

        depth_intv = abs(sample["depths"][1] - sample["depths"][0])

        lim_top_shallowest = np.min(sample["depths"])
        lim_top_deepest = max(
            lim_top_shallowest + depth_intv, np.min(sample["d_bottom"]) - depth_intv
        )
        lim_bot_deepest = np.max(sample["depths"])
        lim_bot_shallowest = min(
            lim_bot_deepest - depth_intv,
            max(np.max(sample["d_turbulence"]), np.min(sample["d_bottom"])),
        )

        if sample["is_upward_facing"]:
            surf_options = sample["d_surface"][sample["d_surface"] > lim_top_shallowest]
            if len(surf_options) == 0:
                opt_top_depth = lim_top_shallowest
            else:
                opt_top_depth = max(lim_top_shallowest, np.min(surf_options))
            opt_bot_depth = lim_bot_deepest
        else:
            opt_top_depth = lim_top_shallowest
            opt_bot_depth = np.max(sample["d_bottom-original"])

        depth_range = abs(opt_bot_depth - opt_top_depth)
        close_dist_grow = self.fraction_close * depth_range
        close_dist_shrink = (
            depth_range * self.fraction_close / (1 + self.fraction_close)
        )

        close_top_shallowest = min(
            lim_top_deepest, max(lim_top_shallowest, opt_top_depth - close_dist_grow)
        )
        close_top_deepest = min(
            lim_top_deepest,
            opt_top_depth + close_dist_shrink,
            np.percentile(sample["d_turbulence"], 25),
        )
        if sample["is_upward_facing"]:
            close_bot_shallowest = max(
                lim_bot_shallowest,
                opt_bot_depth - close_dist_shrink,
            )
        else:
            close_bot_shallowest = max(
                lim_bot_shallowest,
                opt_bot_depth - close_dist_shrink,
                np.percentile(sample["d_bottom-original"], 50),
            )
        close_bot_shallowest = min(lim_bot_deepest, close_bot_shallowest)
        close_bot_deepest = min(
            lim_bot_deepest,
            max(lim_bot_shallowest, opt_bot_depth) + close_dist_grow,
        )

        if (
            close_top_shallowest > close_top_deepest
            or close_bot_shallowest > close_bot_deepest
            or close_top_deepest >= close_bot_shallowest
        ):
            raise ValueError(
                "Nonsensical depth limits:\n"
                "  opt_top_depth        = {:7.3f}\n"
                "  opt_bot_depth        = {:7.3f}\n"
                "  close_top_shallowest = {:7.3f}\n"
                "  close_top_deepest    = {:7.3f}\n"
                "  close_bot_shallowest = {:7.3f}\n"
                "  close_bot_deepest    = {:7.3f}\n".format(
                    opt_top_depth,
                    opt_bot_depth,
                    close_top_shallowest,
                    close_top_deepest,
                    close_bot_shallowest,
                    close_bot_deepest,
                )
            )

        rand_top_shallowest = lim_top_shallowest
        rand_top_deepest = min(
            lim_top_deepest,
            max(
                np.percentile(sample["d_turbulence"], 50),
                opt_top_depth + close_dist_shrink,
            ),
        )
        rand_top_deepest = max(lim_top_shallowest, rand_top_deepest)
        if sample["is_upward_facing"]:
            rand_bot_shallowest = close_bot_shallowest
        else:
            rand_bot_shallowest = max(
                lim_bot_shallowest,
                np.percentile(sample["d_bottom-original"], 50),
            )
        rand_bot_shallowest = min(lim_bot_deepest, rand_bot_shallowest)
        rand_bot_deepest = lim_bot_deepest

        if (
            rand_top_shallowest > rand_top_deepest
            or rand_bot_shallowest > rand_bot_deepest
            or rand_top_deepest >= rand_bot_shallowest
        ):
            raise ValueError(
                "Nonsensical depth limits:\n"
                "  opt_top_depth       = {:7.3f}\n"
                "  opt_bot_depth       = {:7.3f}\n"
                "  rand_top_shallowest = {:7.3f}\n"
                "  rand_top_deepest    = {:7.3f}\n"
                "  rand_bot_shallowest = {:7.3f}\n"
                "  rand_bot_deepest    = {:7.3f}\n".format(
                    opt_top_depth,
                    opt_bot_depth,
                    rand_top_shallowest,
                    rand_top_deepest,
                    rand_bot_shallowest,
                    rand_bot_deepest,
                )
            )

        # Select whether to do a close or fully random crop
        if p < self.p_crop_is_close:
            # Close crop
            top_shallowest = close_top_shallowest
            top_deepest = close_top_deepest
            bot_shallowest = close_bot_shallowest
            bot_deepest = close_bot_deepest
            # Select limits of depth range. Equally likely to expand or contract
            # in each direction.
            if random.random() < 0.5:
                shallowest_depth = random.uniform(top_shallowest, opt_top_depth)
            else:
                shallowest_depth = random.uniform(opt_top_depth, top_deepest)

            if random.random() < 0.5:
                deepest_depth = random.uniform(bot_shallowest, opt_bot_depth)
            else:
                deepest_depth = random.uniform(opt_bot_depth, bot_deepest)
        else:
            # Random crop
            top_shallowest = rand_top_shallowest
            top_deepest = rand_top_deepest
            bot_shallowest = rand_bot_shallowest
            bot_deepest = rand_bot_deepest
            # Select limits of depth range. Either randomly select uniformly
            # across the whole range, or not.
            if (
                sample["is_upward_facing"]
                or random.random() < self.p_nearfield_side_crop
            ):
                shallowest_depth = random.uniform(top_shallowest, top_deepest)
            else:
                shallowest_depth = opt_top_depth
            if (
                not sample["is_upward_facing"]
                or random.random() < self.p_nearfield_side_crop
            ):
                deepest_depth = random.uniform(bot_shallowest, bot_deepest)
            else:
                deepest_depth = opt_bot_depth

        # Crop data
        depth_crop_mask = (shallowest_depth <= sample["depths"]) & (
            sample["depths"] <= deepest_depth
        )

        for key in _fields_1d_depthlike:
            if key in sample:
                sample[key] = sample[key][depth_crop_mask]

        for key in _fields_2d:
            if key in sample:
                sample[key] = sample[key][:, depth_crop_mask]

        return sample


class ColorJitter(object):
    """
    Randomly change the brightness and contrast of a normalized image.

    Note that changes are made inplace.

    Parameters
    ----------
    brightness : float or tuple of float (min, max)
        How much to jitter brightness. ``brightness_factor`` is chosen uniformly
        from ``[-brightness, brightness]`` or the given ``[min, max]``.
        ``brightness_factor`` is then added to the image.
    contrast : float or tuple of float (min, max)
        How much to jitter contrast. ``contrast_factor`` is chosen uniformly from
        ``[max(0, 1 - contrast), 1 + contrast]`` or the given ``[min, max]``.
        Should be non negative numbers.
    """

    def __init__(self, brightness=0, contrast=0):
        self.brightness = self._check_input(
            brightness,
            "brightness",
            center=0,
            bound=(float("-inf"), float("inf")),
            clip_first_on_zero=False,
        )
        self.contrast = self._check_input(contrast, "contrast")

    def _check_input(
        self, value, name, center=1, bound=(0, float("inf")), clip_first_on_zero=True
    ):
        if isinstance(value, (float, int)):
            if value < 0:
                raise ValueError(
                    "If {} is a single number, it must be non negative.".format(name)
                )
            value = [center - value, center + value]
            if clip_first_on_zero:
                value[0] = max(value[0], 0)
        elif isinstance(value, (tuple, list)) and len(value) == 2:
            if not bound[0] <= value[0] <= value[1] <= bound[1]:
                raise ValueError("{} values should be between {}".format(name, bound))
        else:
            raise TypeError(
                "{} should be a single number or a list/tuple with length 2.".format(
                    name
                )
            )

        if value[0] == value[1] == center:
            value = None
        return value

    def __call__(self, sample):
        init_op = random.randint(0, 1)
        for i_op in range(2):
            op_num = (init_op + i_op) % 2
            if op_num == 0 and self.brightness is not None:
                brightness_factor = random.uniform(
                    self.brightness[0], self.brightness[1]
                )
                sample["signals"] += brightness_factor
            elif op_num == 1 and self.contrast is not None:
                contrast_factor = random.uniform(self.contrast[0], self.contrast[1])
                sample["signals"] *= contrast_factor
        return sample

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        format_string += "brightness={0}".format(self.brightness)
        format_string += ", contrast={0})".format(self.contrast)
        format_string += ")"
        return format_string



#!/usr/bin/env python
# coding: utf-8
"""
Convert dataset of CSV exports from Echoview into shards.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import functools
import multiprocessing
import os
import sys
import traceback

from tqdm.autonotebook import tqdm

import echofilter.raw
import echofilter.ui

ROOT_DATA_DIR = echofilter.raw.loader.ROOT_DATA_DIR


def generate_shard(
    transect_pth,
    verbose=False,
    fail_gracefully=True,
    **kwargs,
):
    """
    Shard a single transect.

    Wrapper around echofilter.raw.shardloader.segment_and_shard_transect which
    adds verboseness and graceful failure options.

    Parameters
    ----------
    transect_pth : str
        Relative path to transect.
    verbose : bool, optional
        Whether to print which transect is being processed. Default is ``False``.
    fail_gracefully : bool, optional
        If ``True``, any transect which triggers an errors during processing
        will be printed out, but processing the rest of the transects will
        continue. If ``False``, the process will halt with an error as soon as
        any single transect hits an error. Default is ``True``.
    **kwargs
        See :meth:`echofilter.raw.shardloader.segment_and_shard_transect`.
    """
    if verbose:
        print("Sharding {}".format(transect_pth))
    try:
        echofilter.raw.shardloader.segment_and_shard_transect(
            transect_pth,
            **kwargs,
        )
    except Exception as ex:
        with echofilter.ui.style.error_message():
            if not fail_gracefully:
                raise ex
            print("Error sharding {}".format(transect_pth))
            print("".join(traceback.TracebackException.from_exception(ex).format()))


def generate_shards(
    partition,
    dataset,
    partitioning_version="firstpass",
    progress_bar=False,
    ncores=None,
    verbose=False,
    fail_gracefully=True,
    root_data_dir=ROOT_DATA_DIR,
    **kwargs,
):
    """
    Shard all transections in one partition of a dataset.

    Wrapper around echofilter.raw.shardloader.segment_and_shard_transect which
    adds verboseness and graceful failure options.

    Parameters
    ----------
    partition : str
        Name of the partition to process (``'train'``, ``'validate'``, ``'test'``,
        etc).
    dataset : str
        Name of the dataset to process (``'mobile'``, ``'MinasPassage'``, etc).
    partitioning_version : str, optional
        Name of the partition version to use process. Default is ``'firstpass'``.
    progress_bar : bool, optional
        Whether to output a progress bar using ``tqdm``. Default is ``False``.
    ncores : int, optional
        Number of cores to use for multiprocessing. To disable multiprocessing,
        set to ``1``. Set to ``None`` to use all available cores.
        Default is ``None``.
    verbose : bool, optional
        Whether to print which transect is being processed. Default is ``False``.
    fail_gracefully : bool, optional
        If ``True``, any transect which triggers an errors during processing
        will be printed out, but processing the rest of the transects will
        continue. If ``False``, the process will halt with an error as soon as
        any single transect hits an error. Default is ``True``.
    **kwargs
        See :func:`echofilter.raw.shardloader.segment_and_shard_transect`.
    """
    if verbose:
        print('Getting partition list "{}" for "{}"'.format(partition, dataset))
    transect_pths = echofilter.raw.loader.get_partition_list(
        partition,
        dataset=dataset,
        full_path=False,
        partitioning_version=partitioning_version,
        root_data_dir=root_data_dir,
    )
    if verbose:
        print("Will process {} transects".format(len(transect_pths)))
        print()

    disable_tqdm = not progress_bar
    fn = functools.partial(
        generate_shard,
        dataset=dataset,
        verbose=verbose,
        fail_gracefully=fail_gracefully,
        root_data_dir=root_data_dir,
        **kwargs,
    )
    if ncores == 1:
        for transect_pth in tqdm(
            transect_pths, total=len(transect_pths), disable=disable_tqdm
        ):
            fn(transect_pth)
    else:
        with multiprocessing.Pool(ncores) as pool:
            for _ in tqdm(
                pool.imap_unordered(fn, transect_pths),
                total=len(transect_pths),
                disable=disable_tqdm,
            ):
                pass


def get_parser():
    """
    Build parser for command line interface for generating shards.

    Returns
    -------
    parser : argparse.ArgumentParser
        CLI argument parser for generating shards.
    """
    import argparse

    # Create parser

    prog = os.path.split(sys.argv[0])[1]
    if prog == "__main__.py" or prog == "__main__":
        prog = os.path.split(__file__)[1]
    parser = argparse.ArgumentParser(
        prog=prog,
        description="Generate dataset shards",
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version="%(prog)s {version}".format(version=echofilter.__version__),
    )
    parser.add_argument(
        "partition",
        type=str,
        help="partition to shard",
    )
    parser.add_argument(
        "dataset",
        type=str,
        help="dataset to shard",
    )
    parser.add_argument(
        "--root",
        dest="root_data_dir",
        type=str,
        default=ROOT_DATA_DIR,
        help="root data directory",
    )
    parser.add_argument(
        "--partitioning-version",
        type=str,
        default="firstpass",
        help="partitioning version",
    )
    parser.add_argument(
        "--max-depth",
        type=float,
        default=None,
        help="maximum depth to include in sharded data",
    )
    parser.add_argument(
        "--shard-len",
        type=int,
        default=128,
        help="number of samples in each shard",
    )
    parser.add_argument(
        "--ncores",
        type=int,
        default=None,
        help="number of cores to use (default: all). Set to 1 to disable"
        " multiprocessing.",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="increase verbosity",
    )

    return parser


def _get_parser_sphinx():
    """
    Pre-format parser help for sphinx-argparse processing.
    """
    return echofilter.ui.formatters.format_parser_for_sphinx(get_parser())


def main(args=None):
    """
    Command line interface for generating dataset shards from CSV files.
    """
    parser = get_parser()
    # Parse command line arguments
    args = parser.parse_args(args)

    # Check the input directory exists
    print("Sharding {} partition of {}".format(args.partition, args.dataset))

    # Run command with these arguments
    generate_shards(**vars(args))


if __name__ == "__main__":
    main()



"""
Converting raw data into shards, and loading data from shards.
"""

import os

import numpy as np

from . import loader, manipulate
from .utils import pad1d

ROOT_DATA_DIR = loader.ROOT_DATA_DIR


def segment_and_shard_transect(
    transect_pth,
    dataset="mobile",
    max_depth=None,
    shard_len=128,
    root_data_dir=ROOT_DATA_DIR,
):
    """
    Create a sharded copy of a transect.

    The transect is cut into segments based on recording starts/stops.
    Each segment is split across multiple files (shards) for efficient loading.

    Parameters
    ----------
    transect_pth : str
        Relative path to transect, excluding ``"_Sv_raw.csv"``.
    dataset : str, optional
        Name of dataset. Default is ``"mobile"``.
    max_depth : float or None, optional
        The maximum depth to include in the saved shard. Data corresponding
        to deeper locations is omitted to save on load time and memory when
        the shard is loaded. If ``None``, no cropping is applied.
        Default is ``None``.
    shard_len : int, optional
        Number of timestamp samples to include in each shard. Default is ``128``.
    root_data_dir : str
        Path to root directory where data is located.

    Notes
    -----
    The segments will be written to the directories
    ``<root_data_dir>_sharded/<dataset>/transect_path/<segment>/``
    For the contents of each directory, see ``write_transect_shards``.
    """
    # Define output destination
    root_data_dir = loader.remove_trailing_slash(root_data_dir)
    root_shard_dir = os.path.join(root_data_dir + "_sharded", dataset)

    # Load the data, with mask decomposed into turbulence, bottom, passive,
    # and removed regions.
    transect = manipulate.load_decomposed_transect_mask(
        os.path.join(root_data_dir, dataset, transect_pth)
    )

    segments = manipulate.split_transect(**transect)

    for i_segment, segment in enumerate(segments):
        dirname = os.path.join(root_shard_dir, transect_pth, str(i_segment))
        write_transect_shards(
            dirname, segment, max_depth=max_depth, shard_len=shard_len
        )

    n_segment = i_segment + 1

    # Save segmentation metadata
    with open(os.path.join(root_shard_dir, transect_pth, "n_segment.txt"), "w") as hf:
        print(str(n_segment), file=hf)


def write_transect_shards(dirname, transect, max_depth=None, shard_len=128):
    """
    Create a sharded copy of a transect.

    The transect is cut by timestamp and split across multiple files.

    Parameters
    ----------
    dirname : str
        Path to output directory.
    transect : dict
        Observed values for the transect. Should already be segmented.
    max_depth : float or None, optional
        The maximum depth to include in the saved shard. Data corresponding
        to deeper locations is omitted to save on load time and memory when
        the shard is loaded. If ``None``, no cropping is applied.
        Default is ``None``.
    shard_len : int, optional
        Number of timestamp samples to include in each shard. Default is ``128``.

    Notes
    -----
    The output will be written to the directory ``dirname``, and will contain:

    - a file named ``"shard_size.txt"``, which contains the sharding metadata:
      total number of samples, and shard size;
    - a directory for each shard, named 0, 1, ...
      Each shard directory will contain files:

        - depths.npy
        - timestamps.npy
        - Sv.npy
        - mask.npy
        - turbulence.npy
        - bottom.npy
        - is_passive.npy
        - is_removed.npy
        - is_upward_facing.npy

      which contain pickled numpy dumps of the matrices for each shard.
    """
    # Remove depths which are too deep for us to care about
    if max_depth is not None:
        depth_mask = transect["depths"] <= max_depth
        transect["depths"] = transect["depths"][depth_mask]
        transect["Sv"] = transect["Sv"][:, depth_mask]
        transect["mask"] = transect["mask"][:, depth_mask]

    # Reduce floating point precision for some variables
    for key in ("Sv", "turbulence", "bottom"):
        transect[key] = np.half(transect[key])

    # Ensure is_upward_facing is an array
    transect["is_upward_facing"] = np.array(transect["is_upward_facing"])

    # Prep output directory
    os.makedirs(dirname, exist_ok=True)

    # Save sharding metadata (total number of datapoints, shard size) to
    # make loading from the shards easier
    with open(os.path.join(dirname, "shard_size.txt"), "w") as hf:
        print("{},{}".format(transect["Sv"].shape[0], shard_len), file=hf)

    # Work out where to split the arrays
    indices = range(shard_len, transect["Sv"].shape[0], shard_len)

    # Split the transect into shards
    n_shards = len(indices) + 1
    shards = [{} for _ in range(n_shards)]
    for key in transect:
        if key in ("depths", "is_upward_facing") or not hasattr(
            transect[key], "__len__"
        ):
            for i_shards in range(n_shards):
                shards[i_shards][key] = transect[key]
        else:
            for i_split, split in enumerate(np.split(transect[key], indices)):
                shards[i_split][key] = split

    for shard in shards:
        if shard.keys() != shards[0].keys():
            raise ValueError("Inconsistent split lengths")

    # Save the data for each of the shards
    for i_shard, shard in enumerate(shards):
        fname = os.path.join(dirname, "{}.npz".format(i_shard))
        np.savez_compressed(fname, **shard)


def load_transect_from_shards_abs(
    transect_abs_pth,
    i1=0,
    i2=None,
    pad_mode="edge",
):
    """
    Load transect data from shard files.

    Parameters
    ----------
    transect_abs_pth : str
        Absolute path to transect shard directory.
    i1 : int, optional
        Index of first sample to retrieve. Default is ``0``, the first sample.
    i2 : int, optional
        Index of last sample to retrieve. As-per python convention, the range
        ``i1`` to ``i2`` is inclusive on the left and exclusive on the right, so
        datapoint `i2 - 1` is the right-most datapoint loaded. Default is
        ``None``, which loads everything up to and including to the last sample.
    pad_mode : str, optional
        Padding method for out-of-bounds inputs. Must be supported by
        :meth:`numpy.pad`, such as ``"contast"``, ``"reflect"``, or ``"edge"``. If the mode
        is ``"contast"``, the array will be padded with zeros. Default is "edge".

    Returns
    -------
    dict
        A dictionary with keys:

            - "timestamps" : numpy.ndarray
                Timestamps (in seconds since Unix epoch), for each recording
                timepoint. The number of entries, ``num_timestamps``, is equal
                to `i2 - i1`.
            - "depths" : numpy.ndarray
                Depths from the surface (in metres), with each entry
                corresponding to each column in the ``signals`` data.
            - "Sv" : numpy.ndarray
                Echogram Sv data, shaped (num_timestamps, num_depths).
            - "mask" : numpy.ndarray
                Logical array indicating which datapoints were kept (``True``)
                and which removed (``False``) for the masked Sv output.
                Shaped (num_timestamps, num_depths).
            - "turbulence" : numpy.ndarray
                For each timepoint, the depth of the shallowest datapoint which
                should be included for the mask. Shaped (num_timestamps, ).
            - "bottom" : numpy.ndarray
                For each timepoint, the depth of the deepest datapoint which
                should be included for the mask. Shaped (num_timestamps, ).
            - "is_passive" : numpy.ndarray
                Logical array showing whether a timepoint is of passive data.
                Shaped (num_timestamps, ). All passive recording data should
                be excluded by the mask.
            - "is_removed" : numpy.ndarray
                Logical array showing whether a timepoint is entirely removed
                by the mask. Shaped (num_timestamps, ). Does not include
                periods of passive recording.
            - "is_upward_facing" : bool
                Indicates whether the recording source is located at the
                deepest depth (i.e. the seabed), facing upwards. Otherwise, the
                recording source is at the shallowest depth (i.e. the surface),
                facing downwards.
    """
    # Load the sharding metadata
    with open(os.path.join(transect_abs_pth, "shard_size.txt"), "r") as f:
        n_timestamps, shard_len = f.readline().strip().split(",")
        n_timestamps = int(n_timestamps)
        shard_len = int(shard_len)
    # Set the default value for i2
    if i2 is None:
        i2 = n_timestamps

    # Sanity check
    if i1 > n_timestamps:
        raise ValueError(
            "All requested datapoints out of range: {}, {} > {}".format(
                i1, i2, n_timestamps
            )
        )
    if i2 < 0:
        raise ValueError(
            "All requested datapoints out of range: {}, {} < {}".format(i1, i2, 0)
        )
    # Make indices safe
    i1_ = max(0, i1)
    i2_ = min(i2, n_timestamps)
    # Work out which shards we'll need to load to get this data
    j1 = max(0, int(i1 / shard_len))
    j2 = int(min(i2, n_timestamps - 1) / shard_len)

    transect = {}

    shards = [
        np.load(os.path.join(transect_abs_pth, str(j) + ".npz"), allow_pickle=True)
        for j in range(j1, j2 + 1)
    ]
    # Depths and is_upward_facing should all be the same. Only load one of
    # each of them.
    for key in shards[0].keys():
        if key in ("depths", "is_upward_facing"):
            transect[key] = shards[0][key]
        else:
            broad_data = np.concatenate([shard[key] for shard in shards])
            # Have to trim data down, and pad if requested indices out of range
            transect[key] = pad1d(
                broad_data[(i1_ - j1 * shard_len) : (i2_ - j1 * shard_len)],
                (i1_ - i1, i2 - i2_),
                axis=0,
                mode=pad_mode,
            )

    return transect


def load_transect_from_shards_rel(
    transect_rel_pth,
    i1=0,
    i2=None,
    dataset="mobile",
    segment=0,
    root_data_dir=ROOT_DATA_DIR,
    **kwargs,
):
    """
    Load transect data from shard files.

    Parameters
    ----------
    transect_rel_pth : str
        Relative path to transect.
    i1 : int, optional
        Index of first sample to retrieve. Default is ``0``, the first sample.
    i2 : int, optional
        Index of last sample to retrieve. As-per python convention, the range
        ``i1`` to ``i2`` is inclusive on the left and exclusive on the right, so
        datapoint `i2 - 1` is the right-most datapoint loaded. Default is
        ``None``, which loads everything up to and including to the last sample.
    dataset : str, optional
        Name of dataset. Default is ``"mobile"``.
    segment : int, optional
        Which segment to load. Default is ``0``.
    root_data_dir : str
        Path to root directory where data is located.
    **kwargs
        As per :meth:`load_transect_from_shards_abs`.

    Returns
    -------
    dict
        See :meth:`load_transect_from_shards_abs`.
    """
    root_data_dir = loader.remove_trailing_slash(root_data_dir)
    root_shard_dir = os.path.join(root_data_dir + "_sharded", dataset)
    dirname = os.path.join(root_shard_dir, transect_rel_pth, str(segment))
    return load_transect_from_shards_abs(
        dirname,
        i1=i1,
        i2=i2,
        **kwargs,
    )


def load_transect_segments_from_shards_abs(
    transect_abs_pth,
    segments=None,
):
    """
    Load transect data from shard files.

    Parameters
    ----------
    transect_abs_pth : str
        Absolute path to transect shard segments directory.
    segments : iterable or None
        Which segments to load. If ``None`` (default), all segments are loaded.

    Returns
    -------
    dict
        See :meth:`load_transect_from_shards_abs`.
    """
    if segments is None:
        # Load the segmentation metadata
        with open(os.path.join(transect_abs_pth, "n_segment.txt"), "r") as f:
            n_segment = int(f.readline().strip())
        segments = range(n_segment)

    # Load each segment
    transects = []
    for segment in segments:
        dirname = os.path.join(transect_abs_pth, str(segment))
        transects.append(load_transect_from_shards_abs(dirname))

    # Join the segments together
    return manipulate.join_transect(transects)


def load_transect_segments_from_shards_rel(
    transect_rel_pth,
    dataset="mobile",
    segments=None,
    root_data_dir=ROOT_DATA_DIR,
):
    """
    Load transect data from shard files.

    Parameters
    ----------
    transect_rel_pth : str
        Relative path to transect.
    dataset : str, optional
        Name of dataset. Default is ``"mobile"``.
    segments : iterable or None
        Which segments to load. If ``None`` (default), all segments are loaded.
    root_data_dir : str
        Path to root directory where data is located.
    **kwargs
        As per :meth:`load_transect_from_shards_abs`.

    Returns
    -------
    dict
        See :meth:`load_transect_from_shards_abs`.
    """
    root_data_dir = loader.remove_trailing_slash(root_data_dir)
    root_shard_dir = os.path.join(root_data_dir + "_sharded", dataset)
    dirname = os.path.join(root_shard_dir, transect_rel_pth)
    return load_transect_segments_from_shards_abs(dirname, segments=segments)


# Backwards compatibility
shard_transect = segment_and_shard_transect
load_transect_from_shards = load_transect_from_shards_rel



"""
Manipulating lines and masks contained in Echoview files.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import copy
import os
import warnings

import numpy as np
import scipy.interpolate
import scipy.ndimage

from . import loader, metadata, utils

ROOT_DATA_DIR = loader.ROOT_DATA_DIR


def find_passive_data(signals, n_depth_use=38, threshold=25.0, deviation=None):
    """
    Find segments of Sv recording which correspond to passive recording.

    Parameters
    ----------
    signals : array_like
        Two-dimensional array of Sv values, shaped `[timestamps, depths]`.
    n_depth_use : int, optional
        How many Sv depths to use, starting with the first depths (closest
        to the sounder device). If ``None`` all depths are used. Default is ``38``.
    threshold : float, optional
        Threshold for start/end of passive regions. Default is ``25``.
    deviation : float, optional
        Threshold for start/end of passive regions is ``deviation`` times the
        interquartile-range of the difference between samples at neigbouring
        timestamps. Default is ``None``. Only one of ``threshold`` and ``deviation``
        should be set.

    Returns
    -------
    passive_start : numpy.ndarray
        Indices of rows of ``signals`` at which passive segments start.
    passive_end : numpy.ndarray
        Indices of rows of ``signals`` at which passive segments end.

    Notes
    -----
    Works by looking at the difference between consecutive recordings and
    finding large deviations.
    """
    # Ensure signals is numpy array
    signals = np.asarray(signals)

    if n_depth_use is None:
        n_depth_use = signals.shape[1]

    md = np.median(np.diff(signals[:, :n_depth_use], axis=0), axis=1)

    if threshold is not None and deviation is not None:
        raise ValueError("Only one of `threshold` and `deviation` should be set.")
    if threshold is None:
        if deviation is None:
            raise ValueError("Neither of `threshold` and `deviation` were set.")
        threshold = (np.percentile(md, 75) - np.percentile(md, 25)) * deviation

    threshold_high = threshold
    threshold_low = -threshold
    indices_possible_start = np.nonzero(md < threshold_low)[0]
    indices_possible_end = np.nonzero(md > threshold_high)[0]

    current_index = 0
    indices_passive_start = []
    indices_passive_end = []

    if len(indices_possible_start) == 0 and len(indices_possible_end) == 0:
        return np.array(indices_passive_start), np.array(indices_passive_end)

    if len(indices_possible_start) > 0:
        indices_possible_start += 1

    if len(indices_possible_end) > 0:
        indices_possible_end += 1

    if len(indices_possible_end) > 0 and (
        len(indices_possible_start) == 0
        or indices_possible_end[0] < indices_possible_start[0]
    ):
        indices_passive_start.append(0)
        current_index = indices_possible_end[0]
        indices_passive_end.append(current_index)
        indices_possible_start = indices_possible_start[
            indices_possible_start > current_index
        ]
        indices_possible_end = indices_possible_end[
            indices_possible_end > current_index
        ]

    while len(indices_possible_start) > 0:
        current_index = indices_possible_start[0]
        indices_passive_start.append(current_index)
        baseline = signals[current_index - 1, :n_depth_use]

        # Find first column which returns to the baseline value seen before passive region
        offsets = np.nonzero(
            np.median(baseline - signals[current_index:, :n_depth_use], axis=1)
            < threshold_high
        )[0]
        if len(offsets) == 0:
            current_index = signals.shape[0]
        else:
            current_index += offsets[0]
        indices_passive_end.append(current_index)

        # Remove preceding indices from the list of candidates
        indices_possible_start = indices_possible_start[
            indices_possible_start > current_index
        ]
        indices_possible_end = indices_possible_end[
            indices_possible_end > current_index
        ]

        # Check the start was sufficiently inclusive
        if current_index < signals.shape[0]:
            baseline = signals[current_index, :n_depth_use]
            nonpassives = np.nonzero(
                np.median(baseline - signals[:current_index, :n_depth_use], axis=1)
                < threshold_high
            )[0]
            if len(nonpassives) == 0:
                indices_passive_start[-1] = 0
            else:
                indices_passive_start[-1] = min(
                    indices_passive_start[-1],
                    nonpassives[-1] + 1,
                )

        # Combine with preceding passive segments if they overlap
        while (
            len(indices_passive_start) > 1
            and indices_passive_start[-1] <= indices_passive_end[-2]
        ):
            indices_passive_start = indices_passive_start[:-1]
            indices_passive_end = indices_passive_end[:-2] + indices_passive_end[-1:]

    return np.array(indices_passive_start), np.array(indices_passive_end)


def find_passive_data_v2(
    signals,
    n_depth_use=38,
    threshold_inner=None,
    threshold_init=None,
    deviation=None,
    sigma_depth=0,
    sigma_time=1,
):
    """
    Find segments of Sv recording which correspond to passive recording.

    Parameters
    ----------
    signals : array_like
        Two-dimensional array of Sv values, shaped `[timestamps, depths]`.
    n_depth_use : int, optional
        How many Sv depths to use, starting with the first depths (closest
        to the sounder device). If ``None`` all depths are used. Default is ``38``.
        The median is taken across the depths, after taking the temporal
        derivative.
    threshold_inner : float, optional
        Theshold to apply to the temporal derivative of the signal when
        detected fine-tuned start/end of passive regions.
        Default behaviour is to use a threshold automatically determined using
        ``deviation`` if it is set, and otherwise use a threshold of ``35.0``.
    threshold_init : float, optional
        Theshold to apply during the initial scan of the start/end of passive
        regions, which seeds the fine-tuning search.
        Default behaviour is to use a threshold automatically determined using
        ``deviation`` if it is set, and otherwise use a threshold of ``12.0``.
    deviation : float, optional
        Set ``threshold_inner`` to be ``deviation`` times the standard deviation of
        the temporal derivative of the signal. The standard deviation is
        robustly estimated based on the interquartile range.
        If this is set, ``threshold_inner`` must not be ``None``.
        Default is ``None``
    sigma_depth : float, optional
        Width of kernel for filtering signals across second dimension (depth).
        Default is ``0`` (no filter).
    sigma_time : float, optional
        Width of kernel for filtering signals across second dimension (time).
        Default is ``1``. Set to ``0`` to not filter.

    Returns
    -------
    passive_start : numpy.ndarray
        Indices of rows of ``signals`` at which passive segments start.
    passive_end : numpy.ndarray
        Indices of rows of ``signals`` at which passive segments end.

    Notes
    -----
    Works by looking at the difference between consecutive recordings and
    finding large deviations.
    """
    # Ensure signals is numpy array
    signals = np.asarray(signals)

    if n_depth_use is None:
        n_depth_use = signals.shape[1]

    if sigma_depth > 0:
        signals_smooth = scipy.ndimage.gaussian_filter1d(
            signals.astype(np.float32), sigma_depth, axis=-1
        )
    else:
        signals_smooth = signals

    md_inner = np.median(np.diff(signals_smooth[:, :n_depth_use], axis=0), axis=1)

    if sigma_time > 0:
        signals_init = scipy.ndimage.gaussian_filter1d(
            signals_smooth.astype(np.float32), sigma_time, axis=0
        )
        md_init = np.median(np.diff(signals_init[:, :n_depth_use], axis=0), axis=1)
    else:
        signals_init = signals
        md_init = md_inner

    if threshold_inner is not None and deviation is not None:
        raise ValueError("Only one of `threshold_inner` and `deviation` should be set.")
    if threshold_init is None:
        if deviation is None:
            threshold_init = 12.0
        else:
            threshold_inner = (
                (np.percentile(md_init, 75) - np.percentile(md_init, 25))
                / 1.35
                * deviation
            )
    if threshold_inner is None:
        if deviation is None:
            threshold_inner = 35.0
        else:
            threshold_inner = (
                (np.percentile(md_inner, 75) - np.percentile(md_inner, 25))
                / 1.35
                * deviation
            )

    threshold_high_inner = threshold_inner
    # threshold_low_inner = -threshold_inner
    threshold_high_init = threshold_init
    threshold_low_init = -threshold_init
    indices_possible_start_init = np.nonzero(md_init < threshold_low_init)[0]
    indices_possible_end_init = np.nonzero(md_init > threshold_high_init)[0]

    if len(indices_possible_start_init) == 0 and len(indices_possible_end_init) == 0:
        return np.array([]), np.array([])

    # Fine tune indices without smoothing
    indices_possible_start = []
    indices_possible_end = []

    capture_start = None
    for i, index_p in enumerate(indices_possible_start_init):
        if capture_start is None:
            capture_start = index_p
        if (
            i + 1 >= len(indices_possible_start_init)
            or indices_possible_start_init[i + 1] > index_p + 3
        ):
            # break capture
            capture_end = index_p
            capture = np.arange(capture_start, capture_end + 1)
            indices_possible_start.append(capture[np.argmin(md_init[capture])])
            capture_start = None

    capture_start = None
    for i, index_p in enumerate(indices_possible_end_init):
        if capture_start is None:
            capture_start = index_p
        if (
            i + 1 >= len(indices_possible_end_init)
            or indices_possible_end_init[i + 1] > index_p + 3
        ):
            # break capture
            capture_end = index_p
            capture = np.arange(capture_start, capture_end + 1)
            indices_possible_end.append(capture[np.argmax(md_init[capture])])
            capture_start = None

    indices_possible_start = np.array(indices_possible_start)
    indices_possible_end = np.array(indices_possible_end)

    current_index = 0
    indices_passive_start = []
    indices_passive_end = []

    if len(indices_possible_start) > 0:
        indices_possible_start += 1

    if len(indices_possible_end) > 0:
        indices_possible_end += 1

    if len(indices_possible_end) > 0 and (
        len(indices_possible_start) == 0
        or indices_possible_end[0] < indices_possible_start[0]
    ):
        indices_passive_start.append(0)
        current_index = indices_possible_end[0]
        indices_passive_end.append(current_index)
        indices_possible_start = indices_possible_start[
            indices_possible_start > current_index
        ]
        indices_possible_end = indices_possible_end[
            indices_possible_end > current_index
        ]

    while len(indices_possible_start) > 0:
        current_index = indices_possible_start[0]
        indices_passive_start.append(current_index)
        baseline_index = max(0, current_index - 2)
        baseline = signals[baseline_index, :n_depth_use]

        # Find first column which returns to the baseline value seen before passive region
        offsets = np.nonzero(
            np.median(baseline - signals[current_index:, :n_depth_use], axis=1)
            < threshold_high_inner
        )[0]
        if len(offsets) == 0:
            current_index = signals.shape[0]
        else:
            current_index += offsets[0]
        indices_passive_end.append(current_index)

        # Remove preceding indices from the list of candidates
        indices_possible_start = indices_possible_start[
            indices_possible_start > current_index
        ]
        indices_possible_end = indices_possible_end[
            indices_possible_end > current_index
        ]

        # Check the start was sufficiently inclusive
        if current_index < signals.shape[0]:
            baseline_index = min(signals.shape[0] - 1, current_index + 1)
            baseline = signals[baseline_index, :n_depth_use]
            nonpassives = np.nonzero(
                np.median(baseline - signals[:current_index, :n_depth_use], axis=1)
                < threshold_high_inner
            )[0]
            if len(nonpassives) == 0:
                indices_passive_start[-1] = 0
            else:
                indices_passive_start[-1] = min(
                    indices_passive_start[-1],
                    nonpassives[-1] + 1,
                )

        # Combine with preceding passive segments if they overlap
        while (
            len(indices_passive_start) > 1
            and indices_passive_start[-1] <= indices_passive_end[-2]
        ):
            indices_passive_start = indices_passive_start[:-1]
            indices_passive_end = indices_passive_end[:-2] + indices_passive_end[-1:]

    return np.array(indices_passive_start), np.array(indices_passive_end)


def make_lines_from_mask(mask, depths=None, max_gap_squash=1.0):
    """
    Determine turbulence and bottom lines for a mask array.

    Parameters
    ----------
    mask : array_like
        A two-dimensional logical array, where for each row dimension 1 takes
        the value ``False`` for some unknown continuous stretch at the start and
        end of the column, with ``True`` values between these two masked-out
        regions.
    depths : array_like, optional
        Depth of each sample point along dim 1 of ``mask``. Must be either
        monotonically increasing or monotonically decreasing. Default is the
        index of ``mask``, ``arange(mask.shape[1])``.
    max_gap_squash : float, optional
        Maximum gap to merge together, in metres. Default is ``1.``.

    Returns
    -------
    d_turbulence : numpy.ndarray
        Depth of turbulence line. This is the line of smaller depth which
        separates the ``False`` region of ``mask`` from the central region of
        ``True`` values. (If ``depths`` is monotonically increasing, this is
        for the start of the columns of ``mask``, otherwise it is at the end.)
    d_bottom : numpy.ndarray
        Depth of bottom line. As for ``d_turbulence``, but for the other end of the
        array.
    """
    # Ensure input is an array. Make a copy, so we don't modify the input.
    mask = np.array(mask, copy=True)

    # Autocomplete depth with index.
    if depths is None:
        depths = np.arange(mask.shape[1])
    depths = np.asarray(depths)
    if len(depths) != mask.shape[1]:
        raise ValueError("Length of depths input must match dim 1 of mask.")
    depth_intv = np.median(np.diff(depths))

    # Merge small gaps between masked out data, so the middle is masked out too
    # We merge all gaps smaller than 120 pixels apart.
    max_gap_squash_idx = int(np.round(max_gap_squash / depth_intv))
    mask = utils.squash_gaps(mask, max_gap_squash_idx)

    # Check which depths were removed for each timestamp
    nonremoved_depths = np.tile(depths, (mask.shape[0], 1)).astype("float")
    nonremoved_depths[~mask] = np.nan
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "All-NaN (slice|axis) encountered")
        # Top line is the smallest non-removed depth at each timepoint.
        # We offset by depth_intv / 2 to get a depth in between the last kept
        # value at the top and the first removed value.
        d_turbulence = np.nanmin(nonremoved_depths, axis=1) - depth_intv / 2
        # Bottom line is the largest non-removed depth at each timepoint,
        # offset similarly.
        d_bottom = np.nanmax(nonremoved_depths, axis=1) + depth_intv / 2

    return d_turbulence, d_bottom


def make_lines_from_masked_csv(fname):
    """
    Load a masked csv file and convert its mask to lines.

    Parameters
    ----------
    fname : str
        Path to file containing masked Echoview output data in csv format.

    Returns
    -------
    timestamps : numpy.ndarray
        Sample timestamps.
    d_turbulence : numpy.ndarray
        Depth of turbulence line.
    d_bottom : numpy.ndarray
        Depth of bottom line.
    """
    # Load the masked data
    timestamps, depths, signals_mskd = loader.transect_loader(fname)
    mask = ~np.isnan(signals_mskd)
    d_turbulence, d_bottom = make_lines_from_mask(mask, depths)
    return timestamps, d_turbulence, d_bottom


def write_lines_for_masked_csv(fname_mask, fname_turbulence=None, fname_bottom=None):
    """
    Write turbulence and bottom lines based on masked csv file.

    Parameters
    ----------
    fname_mask : str
        Path to input file containing masked Echoview output data in csv
        format.
    fname_turbulence : str, optional
        Destination of generated turbulence line, written in evl format. If ``None``
        (default), the output name is ``<fname_base>_mask-turbulence.evl``, where
        ``<fname_base>`` is ``fname_mask`` without extension and without any
        occurence of the substrings ``_Sv_raw`` or ``_Sv`` in the base file name.
    fname_bottom : str
        Destination of generated bottom line, written in evl format. If ``None``
        (default), the output name is ``<fname_base>_mask-bottom.evl``.
    """
    if fname_turbulence is None or fname_bottom is None:
        fname_base = os.path.splitext(fname_mask)[0]
        dirname, fname_base = os.path.split(fname_base)
        fname_base = fname_base.replace("_Sv_raw", "").replace("_Sv", "")
        fname_base = os.path.join(dirname, fname_base)
    if fname_turbulence is None:
        fname_turbulence = fname_base + "_mask-turbulence.evl"
    if fname_bottom is None:
        fname_bottom = fname_base + "_mask-bottom.evl"
    # Generate the new lines.
    timestamps, d_turbulence, d_bottom = make_lines_from_masked_csv(fname_mask)
    # Write the new lines to their output files.
    loader.evl_writer(fname_turbulence, timestamps, d_turbulence)
    loader.evl_writer(fname_bottom, timestamps, d_bottom)


def find_nonzero_region_boundaries(v):
    """
    Find the start and end indices for nonzero regions of a vector.

    Parameters
    ----------
    v : array_like
        A vector.

    Returns
    -------
    starts : numpy.ndarray
        Indices for start of regions of nonzero elements in vector ``v``
    ends : numpy.ndarray
        Indices for end of regions of nonzero elements in vector ``v``
        (exclusive).

    Notes
    -----
    For ``i`` in ``range(len(starts))``, the set of values ``v[starts[i]:ends[i]]``
    are nonzero. Values in the range ``v[ends[i]:starts[i+1]]`` are zero.
    """
    v = np.asarray(v)
    v = v != 0
    v = v.astype(np.float)

    starts = np.nonzero(np.diff(v) > 0)[0] + 1
    ends = np.nonzero(np.diff(v) < 0)[0] + 1

    if v[0]:
        starts = np.r_[0, starts]

    if v[-1]:
        ends = np.r_[ends, len(v)]

    return starts, ends


def fixup_lines(
    timestamps,
    depths,
    mask,
    t_turbulence=None,
    d_turbulence=None,
    t_bottom=None,
    d_bottom=None,
):
    """
    Extend existing turbulence/bottom lines based on masked target Sv output.

    Parameters
    ----------
    timestamps : array_like
        Shaped `(num_timestamps, )`.
    depths : array_like
        Shaped `(num_depths, )`.
    mask : array_like
        Boolean array, where ``True`` denotes kept entries.
        Shaped `(num_timestamps, num_depths)`.
    t_turbulence : array_like, optional
        Sampling times for existing turbulence line.
    d_turbulence : array_like, optional
        Depth of existing turbulence line.
    t_bottom : array_like, optional
        Sampling times for existing bottom line.
    d_bottom : array_like, optional
        Depth of existing bottom line.

    Returns
    -------
    d_turbulence_new : numpy.ndarray
        Depth of new turbulence line.
    d_bottom_new : numpy.ndarray
        Depth of new bottom line.
    """
    # Handle different sampling grids
    if d_turbulence is not None:
        if t_turbulence is None:
            raise ValueError(
                "t_turbulence must be provided if d_turbulence is provided"
            )
        d_turbulence = np.interp(timestamps, t_turbulence, d_turbulence)

    if d_bottom is not None:
        if t_bottom is None:
            raise ValueError("t_bottom must be provided if d_bottom is provided")
        d_bottom = np.interp(timestamps, t_bottom, d_bottom)

    # Generate fresh lines corresponding to said mask
    d_turbulence_new, d_bottom_new = make_lines_from_mask(mask, depths)

    # Ensure nans in the lines are replaced with prior values, if possible
    li = np.isnan(d_turbulence_new)
    if d_turbulence is not None:
        d_turbulence_new[li] = d_turbulence[li]
    elif np.any(~li):
        d_turbulence_new[li] = np.interp(
            timestamps[li],
            timestamps[~li],
            d_turbulence_new[~li],
        )
    li = np.isnan(d_bottom_new)
    if d_bottom is not None:
        d_bottom_new[li] = d_bottom[li]
    elif np.any(~li):
        d_bottom_new[li] = np.interp(
            timestamps[li],
            timestamps[~li],
            d_bottom_new[~li],
        )

    # Ensure that the lines cover at least as much material as they did before
    if d_turbulence is not None:
        d_turbulence_new = np.maximum(d_turbulence, d_turbulence_new)
    if d_bottom is not None:
        d_bottom_new = np.minimum(d_bottom, d_bottom_new)

    # This mask can't handle regions where all the data was removed.
    # Find those and replace them with the original lines, if they were
    # provided. If they weren't, interpolate to fill the holes (if there is
    # something to interpolate).
    all_removed = ~np.any(mask, axis=1)
    any_all_removed = np.any(all_removed)
    everything_removed = np.all(all_removed)
    if not any_all_removed:
        pass
    elif d_turbulence is not None:
        d_turbulence_new[all_removed] = d_turbulence[all_removed]
    elif ~everything_removed:
        d_turbulence_new[all_removed] = np.interp(
            timestamps[all_removed],
            timestamps[~all_removed],
            d_turbulence_new[~all_removed],
        )
    else:
        d_turbulence_new[all_removed] = np.nan

    if not any_all_removed:
        pass
    elif d_bottom is not None:
        d_bottom_new[all_removed] = d_bottom[all_removed]
    elif ~everything_removed:
        d_bottom_new[all_removed] = np.interp(
            timestamps[all_removed],
            timestamps[~all_removed],
            d_bottom_new[~all_removed],
        )
    else:
        d_bottom_new[all_removed] = np.nan

    return d_turbulence_new, d_bottom_new


def remove_anomalies_1d(
    signal, thr=5, thr2=4, kernel=201, kernel2=31, return_filtered=False
):
    """
    Remove anomalies from a temporal signal.

    Apply a median filter to the data, and replaces datapoints which
    deviate from the median filtered signal by more than some threshold
    with the median filtered data. This process is repeated until no
    datapoints deviate from the filtered line by more than the threshold.

    Parameters
    ----------
    signal : array_like
        The signal to filter.
    thr : float, optional
        The initial threshold will be ``thr`` times the standard deviation of the
        residuals. The standard deviation is robustly estimated from the
        interquartile range. Default is ``5``.
    thr2 : float, optional
        The threshold for repeated iterations will be ``thr2`` times the standard
        deviation of the remaining residuals. The standard deviation is
        robustly estimated from interdecile range. Default is ``4``.
    kernel : int, optional
        The kernel size for the initial median filter. Default is ``201``.
    kernel2 : int, optional
        The kernel size for subsequent median filters. Default is ``31``.
    return_filtered : bool, optional
        If ``True``, the median filtered signal is also returned.
        Default is ``False``.

    Returns
    -------
    signal : numpy.ndarray like signal
        The input signal with anomalies replaced with median values.
    is_replaced : bool numpy.ndarray shaped like signal
        Indicator for which datapoints were replaced.
    filtered : numpy.ndarray like signal, optional
        The final median filtered signal. Returned if ``return_filtered=True``.

    See Also
    --------
    echofilter.raw.utils.medfilt1d
    """
    signal = np.copy(signal)

    # Median filtering, with reflection padding
    filtered = utils.medfilt1d(signal, kernel)
    # Measure the residual between the original and median filtered signal
    residual = signal - filtered
    # Replace datapoints more than thr sigma away from the median filter
    # with the filtered signal. We use a robust estimate of the standard
    # deviation, using the central 50% of datapoints.
    stdev = np.diff(np.percentile(residual, [25, 75])).item() / 1.35
    is_replaced = np.abs(residual) > thr * stdev
    signal[is_replaced] = filtered[is_replaced]

    # Filter again, with a narrower kernel but tighter threshold
    while True:
        filtered = utils.medfilt1d(signal, kernel2)
        # Mesure new residual
        residual = signal - filtered
        # Make sure to only include original data points when determining
        # the standard deviation. We use the interdecile range.
        stdev = np.diff(np.percentile(residual[~is_replaced], [10, 90])).item() / 2.56
        is_replaced_now = np.abs(residual) > thr2 * stdev
        is_replaced |= is_replaced_now
        signal[is_replaced] = filtered[is_replaced]
        # We are done when no more datapoints had to be replaced
        if not np.any(is_replaced_now):
            break

    if return_filtered:
        return signal, is_replaced, filtered
    return signal, is_replaced


def fix_surface_line(timestamps, d_surface, is_passive):
    """
    Fix anomalies in the surface line.

    Parameters
    ----------
    timestamps : array_like sized (N, )
        Timestamps for each ping.
    d_surface : array_like sized (N, )
        Surface line depths.
    is_passive : array_like sized (N, )
        Indicator for passive data. Values for the surface line during passive
        data collection will not be used.

    Returns
    -------
    fixed_surface : numpy.ndarray
        Surface line depths, with anomalies replaced with median filtered
        values and passive data replaced with linear interpolation.
        Has the same size and dtype as ``d_surface``.
    is_replaced : boolean numpy.ndarray sized (N, )
        Indicates which datapoints were replaced. Note that passive data is
        always replaced and is marked as such.
    """
    # Ensure is_passive is a boolean array
    is_passive = is_passive > 0.5

    # Initialise
    out_timestamps = []
    out_surface = []
    out_is_replaced = []
    out_filtered = []

    # Process each segment separately. We divide up into segments after
    # removing any passive data.
    for segment in split_transect(
        timestamps[~is_passive], d_surface=d_surface[~is_passive]
    ):
        fixed, is_replaced, filtered = remove_anomalies_1d(
            segment["d_surface"], return_filtered=True
        )
        out_timestamps.append(segment["d_surface"])
        out_surface.append(fixed)
        out_is_replaced.append(is_replaced)
        out_filtered.append(filtered)

    # Merge segments into a single array for each output
    out_timestamps = np.concatenate(out_timestamps)
    out_surface = np.concatenate(out_surface)
    out_is_replaced = np.concatenate(out_is_replaced)
    out_filtered = np.concatenate(out_filtered)

    # Add back datapoints during passive by interpolating the filtered signal
    # over time
    fixed_surface = np.zeros_like(d_surface)
    fixed_surface[~is_passive] = out_surface
    fixed_surface[is_passive] = np.interp(
        timestamps[is_passive], timestamps[~is_passive], out_filtered
    )
    # Include indication that passive data was always replaced as well as other
    # datapoints
    is_replaced = np.ones_like(is_passive)
    is_replaced[~is_passive] = out_is_replaced

    return fixed_surface, is_replaced


def load_decomposed_transect_mask(sample_path):
    """
    Load a raw and masked transect and decompose the mask.

    The mask is decomposed into turbulence and bottom lines, and passive and
    removed regions.

    Parameters
    ----------
    sample_path : str
        Path to sample, without extension. The raw data should be located at
        ``sample_path + "_Sv_raw.csv"``.

    Returns
    -------
    dict
        A dictionary with keys:

            - "timestamps" : numpy.ndarray
                Timestamps (in seconds since Unix epoch), for each recording
                timepoint.
            - "depths" : numpy.ndarray
                Depths from the surface (in metres), with each entry
                corresponding to each column in the ``signals`` data.
            - "Sv" : numpy.ndarray
                Echogram Sv data, shaped (num_timestamps, num_depths).
            - "mask" : numpy.ndarray
                Logical array indicating which datapoints were kept (``True``)
                and which removed (``False``) for the masked Sv output.
                Shaped (num_timestamps, num_depths).
            - "turbulence" : numpy.ndarray
                For each timepoint, the depth of the shallowest datapoint which
                should be included for the mask. Shaped (num_timestamps, ).
            - "bottom" : numpy.ndarray
                For each timepoint, the depth of the deepest datapoint which
                should be included for the mask. Shaped (num_timestamps, ).
            - "is_passive" : numpy.ndarray
                Logical array showing whether a timepoint is of passive data.
                Shaped (num_timestamps, ). All passive recording data should
                be excluded by the mask.
            - "is_removed" : numpy.ndarray
                Logical array showing whether a timepoint is entirely removed
                by the mask. Shaped (num_timestamps, ). Does not include
                periods of passive recording.
            - "is_upward_facing" : bool
                Indicates whether the recording source is located at the
                deepest depth (i.e. the seabed), facing upwards. Otherwise, the
                recording source is at the shallowest depth (i.e. the surface),
                facing downwards.
    """
    # Load raw data
    fname_raw = os.path.join(sample_path + "_Sv_raw.csv")
    fname_masked = os.path.join(sample_path + "_Sv.csv")

    ts_raw, depths_raw, signals_raw = loader.transect_loader(fname_raw)
    ts_mskd, depths_mskd, signals_mskd = loader.transect_loader(fname_masked)
    mask = ~np.isnan(signals_mskd)

    # Determine whether depths are ascending or descending
    is_upward_facing = depths_raw[-1] < depths_raw[0]

    fname_turbulence1 = sample_path + "_turbulence.evl"
    fname_turbulence2 = sample_path + "_air.evl"
    fname_bottom = sample_path + "_bottom.evl"
    fname_surface = sample_path + "_surface.evl"

    if os.path.isfile(fname_turbulence1):
        fname_turbulence = fname_turbulence1
        if os.path.isfile(fname_turbulence2):
            raise ValueError(
                "Only one of {} and {} should exist.".format(
                    fname_turbulence1, fname_turbulence2
                )
            )
    elif os.path.isfile(fname_turbulence2):
        fname_turbulence = fname_turbulence2
    else:
        raise ValueError(
            "Neither {} nor {} were found.".format(fname_turbulence1, fname_turbulence2)
        )
    t_turbulence, d_turbulence = loader.evl_loader(fname_turbulence)

    if os.path.isfile(fname_bottom):
        t_bottom, d_bottom = loader.evl_loader(fname_bottom)
    elif not is_upward_facing:
        raise ValueError(
            "Expected {} to exist when transect is downfacing.".format(fname_bottom)
        )
    else:
        # Default bottom depth is the bottom of the field of view
        t_bottom = ts_raw
        d_bottom = np.ones_like(ts_raw) * np.max(depths_raw)

    if os.path.isfile(fname_surface):
        t_surface, d_surface = loader.evl_loader(fname_surface)
    elif is_upward_facing:
        raise ValueError(
            "Expected {} to exist when transect is upfacing.".format(fname_surface)
        )
    else:
        # Default surface depth of 0m for downward facing data
        t_surface = ts_raw
        d_surface = np.zeros_like(ts_raw)

    # Generate new lines from mask
    d_turbulence_new, d_bottom_new = fixup_lines(
        ts_mskd,
        depths_mskd,
        mask,
        t_turbulence=t_turbulence,
        d_turbulence=d_turbulence,
        t_bottom=t_bottom,
        d_bottom=d_bottom,
    )

    def tidy_up_line(t, d):
        if d is None:
            return np.nan * np.ones_like(ts_raw)
        is_usable = np.isfinite(d)
        if np.sum(is_usable) > 0:
            t = t[is_usable]
            d = d[is_usable]
        return np.interp(ts_raw, t, d)

    # Mask and data derived from it is sampled at the correct timestamps and
    # depths for the raw data. It should be, but might not be if either of the
    # CSV files contained invalid data.
    if (
        len(ts_raw) != len(ts_mskd)
        or len(depths_raw) != len(depths_mskd)
        or not np.allclose(ts_raw, ts_mskd)
        or not np.allclose(depths_raw, depths_mskd)
    ):
        # Interpolate depth lines to timestamps used for raw data
        d_turbulence_new = tidy_up_line(ts_mskd, d_turbulence_new)
        d_bottom_new = tidy_up_line(ts_mskd, d_bottom_new)
        # Interpolate mask
        if is_upward_facing:
            mask = scipy.interpolate.RectBivariateSpline(
                ts_mskd,
                depths_mskd[::-1],
                mask[:, ::-1].astype(np.float),
            )(ts_raw, depths_raw[::-1])[:, ::-1]
        else:
            mask = scipy.interpolate.RectBivariateSpline(
                ts_mskd,
                depths_mskd,
                mask.astype(np.float),
            )(ts_raw, depths_raw)
        # Binarise
        mask = mask > 0.5

    # Find location of passive data
    passive_edges = metadata.recall_passive_edges(sample_path, ts_raw)
    if passive_edges[0] is not None:
        passive_starts, passive_ends = passive_edges
    else:
        passive_starts, passive_ends = find_passive_data(signals_raw)
    # Determine whether each timestamp is for a period of passive recording
    is_passive = np.zeros(ts_raw.shape, dtype=bool)
    for pass_start, pass_end in zip(passive_starts, passive_ends):
        is_passive[pass_start:pass_end] = True

    # Determine whether each timestamp is for recording which was completely
    # removed from analysis (but not because it is passive recording)
    if np.all(mask):
        print("No data points were masked out in {}".format(fname_masked))
        # Use lines to create a mask
        ddepths = np.broadcast_to(depths_mskd, signals_mskd.shape)
        mask[ddepths < np.expand_dims(d_turbulence_new, -1)] = 0
        mask[ddepths > np.expand_dims(d_bottom_new, -1)] = 0
        mask[is_passive] = 0
    allnan = np.all(~mask, axis=1)

    # Timestamp is entirely removed if everything is nan and it isn't passive
    is_removed_raw = allnan & ~is_passive
    # But we don't want to include removed segments which are marked as
    # removed just because the lines crossed each other.
    r_starts_raw, r_ends_raw = utils.get_indicator_onoffsets(is_removed_raw)
    r_starts = []
    r_ends = []
    is_removed = np.zeros_like(is_removed_raw)
    for r_start, r_end in zip(r_starts_raw, r_ends_raw):
        # Check how many points in the fully removed region don't have
        # overlapping turbulence and bottom lines
        n_without_overlap = np.sum(
            d_turbulence_new[r_start : r_end + 1] < d_bottom_new[r_start : r_end + 1]
        )
        if n_without_overlap == 0:
            # Region is removed only by virtue of the lines crossing; we
            # don't include this
            continue
        if r_end - r_start >= 4 and n_without_overlap <= 2:
            # We expect more than just the edges of the boundary for the region
            # to have uncrossed lines
            continue
        r_starts.append(r_start)
        r_ends.append(r_end)
        is_removed[r_start : r_end + 1] = 1

    # Ensure depth is always increasing (which corresponds to descending from
    # the air down the water column)
    if is_upward_facing:
        depths_raw = depths_raw[::-1].copy()
        signals_raw = signals_raw[:, ::-1].copy()
        mask = mask[:, ::-1].copy()

    # Offset by a small amount to catch pixels on the edge of the line
    depth_intv = abs(depths_raw[1] - depths_raw[0])
    if d_turbulence is not None:
        d_turbulence += depth_intv / 4
    if d_bottom is not None:
        d_bottom -= depth_intv / 4

    d_turbulence = tidy_up_line(t_turbulence, d_turbulence)
    d_bottom = tidy_up_line(t_bottom, d_bottom)
    d_surface = tidy_up_line(t_surface, d_surface)

    # Fix up surface line, removing anomalies
    d_surface, is_surrogate_surface = fix_surface_line(ts_raw, d_surface, is_passive)

    # Make a mask indicating left-over patches. This is 0 everywhere,
    # except 1s wherever pixels in the overall mask are removed for
    # reasons not explained by the turbulence and bottom lines, and is_passive and
    # is_removed indicators.
    mask_patches = ~mask
    mask_patches[is_passive] = 0
    mask_patches[is_removed > 0.5] = 0
    mask_patches_og = mask_patches.copy()
    mask_patches_ntob = mask_patches.copy()
    ddepths = np.broadcast_to(depths_raw, signals_raw.shape)
    mask_patches[ddepths <= np.expand_dims(d_turbulence_new, -1)] = 0
    mask_patches[ddepths >= np.expand_dims(d_bottom_new, -1)] = 0
    mask_patches_og[ddepths <= np.expand_dims(d_turbulence, -1)] = 0
    mask_patches_og[ddepths >= np.expand_dims(d_bottom, -1)] = 0
    mask_patches_ntob[ddepths <= np.expand_dims(d_turbulence_new, -1)] = 0
    mask_patches_ntob[ddepths >= np.expand_dims(d_bottom, -1)] = 0
    # Remove trivial mask patches. If the pixel above and below are both empty,
    # delete a mask with a height of only one-pixel.
    mask_patches[
        ~(
            np.concatenate(
                (
                    mask_patches[:, 2:],
                    np.ones((mask_patches.shape[0], 2), dtype="bool"),
                ),
                axis=-1,
            )
            | np.concatenate(
                (
                    np.ones((mask_patches.shape[0], 2), dtype="bool"),
                    mask_patches[:, :-2],
                ),
                axis=-1,
            )
        )
    ] = 0
    mask_patches_og[
        ~(
            np.concatenate(
                (
                    mask_patches_og[:, 2:],
                    np.ones((mask_patches_og.shape[0], 2), dtype="bool"),
                ),
                axis=-1,
            )
            | np.concatenate(
                (
                    np.ones((mask_patches_og.shape[0], 2), dtype="bool"),
                    mask_patches_og[:, :-2],
                ),
                axis=-1,
            )
        )
    ] = 0
    mask_patches_ntob[
        ~(
            np.concatenate(
                (
                    mask_patches_ntob[:, 2:],
                    np.ones((mask_patches_ntob.shape[0], 2), dtype="bool"),
                ),
                axis=-1,
            )
            | np.concatenate(
                (
                    np.ones((mask_patches_ntob.shape[0], 2), dtype="bool"),
                    mask_patches_ntob[:, :-2],
                ),
                axis=-1,
            )
        )
    ] = 0

    # Collate transect as a dictionary
    transect = {}
    transect["timestamps"] = ts_raw
    transect["depths"] = depths_raw
    transect["Sv"] = signals_raw
    transect["mask"] = mask
    transect["mask_patches"] = mask_patches
    transect["mask_patches-original"] = mask_patches_og
    transect["mask_patches-ntob"] = mask_patches_ntob
    transect["turbulence"] = d_turbulence_new
    transect["turbulence-original"] = d_turbulence
    transect["bottom"] = d_bottom_new
    transect["bottom-original"] = d_bottom
    transect["surface"] = d_surface
    transect["is_surrogate_surface"] = is_surrogate_surface
    transect["is_passive"] = is_passive
    transect["is_removed"] = is_removed
    transect["is_upward_facing"] = is_upward_facing

    return transect


def split_transect(
    timestamps=None,
    threshold=20,
    percentile=97.5,
    max_length=-1,
    pad_length=32,
    pad_on="max",
    **transect,
):
    """
    Split a transect into segments each containing contiguous recordings.

    Parameters
    ----------
    timestamps : array_like
        A 1-d array containing the timestamp at which each recording was
        measured. The sampling is assumed to high-frequency with
        occassional gaps.
    threshold : int, optional
        Threshold for splitting timestamps into segments. Any timepoints
        further apart than ``threshold`` times the ``percentile`` percentile of the
        difference between timepoints will be split apart into new segments.
        Default is ``20``.
    percentile : float, optional
        The percentile at which to sample the timestamp intervals to establish
        a baseline typical interval. Default is ``97.5``.
    max_length : int, default=-1
        Maximum length of each segment.
        Set to ``0`` or ``-1`` to disable (default).
    pad_length : int, default=32
        Amount of overlap between the segments. Set to ``0`` to disable.
    pad_on : {"max", "thr", "all", "none"}, default="max"
        Apply overlap padding when the transect is split due to either the total
        length exceeding the maximum (``"max"``), the time delta exceeding the
        threshold (``"thr"``), or both (``"all"``).
    **kwargs
        Arbitrary additional transect variables, which will be split into
        segments as appropriate in accordance with ``timestamps``.

    Yields
    ------
    dict
        Containing segmented data, key/value pairs as per given in ``**kwargs``
        in addition to ``timestamps``.
    """
    if timestamps is None:
        raise ValueError("The `timestamps` argument is required.")

    if pad_on not in {"max", "thr", "all", "none"}:
        raise ValueError("Unsupported pad_on values: {}".format(pad_on))

    if max_length is None or max_length < 0:
        max_length = 0

    # Fix break points due to time delta threshold
    dt = np.diff(timestamps)
    break_indices = np.where(dt > np.percentile(dt, percentile) * threshold)[0]
    if len(break_indices) > 0:
        break_indices += 1

    # Add the start and end indices
    break_indices = np.r_[0, break_indices, len(timestamps)]

    # Check for breaks needed due to max length
    break_indices2 = []
    break_from_max_pre = []
    break_from_max_post = []
    for i in range(len(break_indices) - 1):
        segment_len = break_indices[i + 1] - break_indices[i]
        if not max_length or segment_len <= max_length:
            break_indices2.append(break_indices[i])
            break_from_max_pre.append(False)
            break_from_max_post.append(False)
            continue
        subseg = np.linspace(
            break_indices[i],
            break_indices[i + 1],
            1 + int(np.ceil(segment_len / max_length)),
        )
        subseg = np.round(subseg).astype(int)
        for j, subseg_idx in enumerate(subseg[:-1]):
            break_indices2.append(subseg_idx)
            break_from_max_pre.append(j > 0)
            break_from_max_post.append(j < len(subseg) - 2)

    break_indices2.append(len(timestamps))

    for seg_start, seg_end, seg_max_pre, seg_max_post in zip(
        break_indices2[:-1], break_indices2[1:], break_from_max_pre, break_from_max_post
    ):
        segment = {}
        # Determine whether to pad the start/end, and record this
        segment["_pad_start"] = 0
        segment["_pad_end"] = 0
        if (
            pad_on == "all"
            or (pad_on == "max" and seg_max_pre)
            or (pad_on == "thr" and not seg_max_pre)
        ):
            segment["_pad_start"] = pad_length
        if (
            pad_on == "all"
            or (pad_on == "max" and seg_max_post)
            or (pad_on == "thr" and not seg_max_post)
        ):
            segment["_pad_end"] = pad_length
        if seg_start == 0:
            segment["_pad_start"] = 0
        if seg_end == len(timestamps):
            segment["_pad_end"] = 0
        # The padding effects the start and end of the extracted indices
        seg_start -= segment["_pad_start"]
        seg_end += segment["_pad_end"]
        segment["timestamps"] = timestamps[seg_start:seg_end]
        for key in transect:
            if key in ("depths",) or np.asarray(transect[key]).size <= 1:
                segment[key] = copy.deepcopy(transect[key])
            else:
                segment[key] = transect[key][seg_start:seg_end]
        yield segment


def join_transect(transects):
    """
    Join segmented transects together into a single dictionary.

    Parameters
    ----------
    transects : iterable of dict
        Transect segments, each with the same fields and compatible shapes.

    Yields
    ------
    dict
        Transect data.
    """
    non_timelike_dims = ["depths"]
    non_output_keys = {"_pad_start", "_pad_end"}

    for i, transect in enumerate(transects):
        if "depths" not in transect:
            raise ValueError(
                "'depths' is a required field, not found in transect {}".format(i)
            )
        if i == 0:
            output = {k: [] for k in transect if k not in non_output_keys}
            output["depths"] = transect["depths"]
            for key in transect:
                if key in non_output_keys:
                    continue
                if np.asarray(transect[key]).size <= 1:
                    output[key] = transect[key]
                    non_timelike_dims.append(key)
        if not np.allclose(output["depths"], transect["depths"]):
            raise ValueError("'depths' must be the same for all segments.")

        scan_start = 0
        scan_end = 0
        if "_pad_start" in transect:
            scan_start += transect.pop("_pad_start")
        if "_pad_end" in transect:
            scan_end -= transect.pop("_pad_end")
        if scan_end == 0:
            scan_end = None

        if transect.keys() != output.keys():
            raise ValueError("Keys mismatch.")

        for key in output:
            if key in non_output_keys:
                continue
            if key in non_timelike_dims:
                continue
            output[key].append(transect[key][scan_start:scan_end])

    for key in output:
        if key in non_timelike_dims:
            continue
        output[key] = np.concatenate(output[key], axis=0)

    return output


def pad_transect(transect, pad=32, pad_mode="reflect", previous_padding="diff"):
    """
    Pad a transect in the timestamps dimension (axis 0).

    Parameters
    ----------
    transect : dict
        A dictionary of transect data.

    pad : int, default=32
        Amount of padding to add.

    pad_mode : str, default="reflect"
        Padding method for out-of-bounds inputs. Must be supported by
        :meth:`numpy.pad`, such as ``"contast"``, ``"reflect"``, or ``"edge"``.
        If the mode is ``"contast"``, the array will be padded with zeros.

    previous_padding : {"diff", "add", "noop"}, default="diff"
        How to handle this padding if the transect has already been padded.

            ``"diff"``
                Extend the padding up to the target ``pad`` value.
            ``"add"``
                Add this padding irrespective of pre-existing padding.
            ``"noop"``
                Don't add any new padding if previously padded.

    Returns
    -------
    transect : dict
        Like input ``transect``, but with all time-like dimensions extended
        with padding and fields ``"_pad_start"`` and ``"_pad_end"`` changed
        to indicate the total padding (including any pre-existing padding).
    """
    pad_shape = [pad, pad]

    if previous_padding == "diff":
        # Only extend existing padding up to target amount of padding
        pad_shape[0] -= transect.get("_pad_start", 0)
        pad_shape[1] -= transect.get("_pad_end", 0)
        pad_shape[0] = max(0, pad_shape[0])
        pad_shape[1] = max(0, pad_shape[1])
    elif previous_padding == "noop":
        # Don't do anything to edges already padded
        if transect.get("_pad_start", 0) != 0:
            pad_shape[0] = 0
        if transect.get("_pad_end", 0) != 0:
            pad_shape[1] = 0
    elif previous_padding == "add":
        # Add new padding in addition to previous padding
        pass
    else:
        raise ValueError(
            "Unfamiliar previous_padding handling mode: {}".format(previous_padding)
        )

    if pad_shape[0] == pad_shape[1] == 0:
        # Nothing to do
        return transect

    if "_pad_start" in transect:
        transect["_pad_start"] += pad_shape[0]
    else:
        transect["_pad_start"] = pad_shape[0]
    if "_pad_end" in transect:
        transect["_pad_end"] += pad_shape[1]
    else:
        transect["_pad_end"] = pad_shape[1]

    for key in transect.keys():
        if key in ("depths", "is_upward_facing") or not hasattr(
            transect[key], "__len__"
        ):
            continue
        transect[key] = utils.pad1d(transect[key], pad_shape, axis=0, mode=pad_mode)

    return transect



"""
U-Net model.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import functools

import torch
from torch import nn

from . import modules


class Down(nn.Module):
    """
    Downscaling layer, downsampling by a factor of two in one or more dimensions.
    """

    def __init__(self, mode="max", compress_dims=True):
        super(Down, self).__init__()

        compress_dims = modules.utils._pair(compress_dims)
        kernel_sizes = modules.utils._pair(
            2 if compress_dim else 1 for compress_dim in compress_dims
        )

        if mode == "max":
            self.pool = nn.MaxPool2d(kernel_sizes)
        elif mode == "avg":
            self.pool = nn.AvgPool2d(kernel_sizes)
        else:
            raise ValueError("Unsupported pooling method: {}".format(mode))

    def forward(self, x):
        return self.pool(x)


class Up(nn.Module):
    """
    Upscaling layer, upsampling by a factor of two in one or more dimensions.
    """

    def __init__(self, in_channels=None, up_dims=True, mode="bilinear"):
        super(Up, self).__init__()

        up_dims = modules.utils._pair(up_dims)
        kernel_sizes = modules.utils._pair(2 if up_dim else 1 for up_dim in up_dims)

        # If conv mode, use a transposed convolution to increase the size
        # Otherwise, use one of the nn.Upsample modes:
        # {"nearest", "linear", "bilinear", "bicubic"}
        if "conv" in mode:
            if in_channels is None:
                raise ValueError(
                    "Number of channels must be provided if upscaling with "
                    "transposed convolution."
                )
            self.up = nn.ConvTranspose2d(
                in_channels,
                in_channels,
                kernel_size=kernel_sizes,
                stride=kernel_sizes,
            )
        else:
            self.up = nn.Upsample(
                scale_factor=kernel_sizes, mode=mode, align_corners=True
            )

    def forward(self, x):
        return self.up(x)


class UNetBlock(nn.Module):
    """
    Create a (cascading set of) UNet block(s).

    Each block performs the steps:
        - Store input to be used in skip connection
        - Down step
        - Horizontal block
        - <Recursion>
        - Up step
        - Concatenate with skip connection
        - Horizontal block

    Where <Recursion> is a call generating a child UNetBlock instance.

    Parameters
    ----------
    in_channels : int
        Number of input channels to this block.
    horizontal_block_factory : callable
        A :class:`torch.nn.Module` constructor or function which returns a
        block of layers. The resulting module must accept ``in_channels`` and
        ``out_channels`` as its first two arguments.
    n_block : int, optional
        The number of nested UNetBlocks to use. Default is ``1`` (no nesting).
    block_expansion_factor : int or float, optional
        Expansion factor for the number of channels between nested UNetBlocks.
        Default is ``2``.
    expand_only_on_down : bool, optional
        Whether to exand the number of channels only when one of the spatial
        dimensions is compressed. Default is ``False``.
    blocks_per_downsample : int or sequence, optional
        How many blocks to include between each downsample operation. This can
        be a tuple of values for each spatial dimension, or an int which
        uses the same value for each spatial dimension. Default is ``1``.
    blocks_before_first_downsample : int or sequence, optional
        How many blocks to include before the first spatial downsampling
        occurs. Default is ``1``.
    always_include_skip_connection : bool, optional
        If ``True``, a skip connection is included even if no dimensions were
        downsampled in this block. Default is ``True``.
    deepest_inner : {callable, "horizontal_block", "identity", None}, optional
        A layer which should be applied at the deepest part of the network,
        before the first upsampling step. The parameter should either be a
        pre-instantiated layer, or the string ``"horizontal_block"``, to indicate
        an additional block as generated by the ``horizontal_block_factory``.
        If it is the string ``"identity"`` or ``None`` (default), no additional
        layer is included at the deepest point before upsampling begins.
    downsampling_modes : {"max", "avg", "stride"} or sequence, optional
        The downsampling mode to use. If this is a string, the same
        downsampling mode is used for every downsampling step. If it is
        a sequence, it should contain a string for each downsampling step.
        If the input sequence is too short, the final value will be used
        for all remaining downsampling steps. Default is ``"max"``.
    upsampling_modes : str or sequence, optional
        The upsampling mode to use. If this is a string, it must be ``"conv"``,
        or something supported by :class:`torch.nn.Upsample`; the same
        upsampling mode is used for every upsampling step. If it is
        a sequence, it should contain a string for each upsampling step.
        If the input sequence is too short, the final value will be used
        for all remaining upsampling steps. Default is ``"bilinear"``.
    _i_block : int, optional
        The current block number. Used internally to track recursion.
        Default is ``0``.
    _i_down : int, optional
        Used internally to track downsampling depth. Default is ``0``.

    Notes
    -----
    This class is defined recursively, and will instantiate itself as its own
    child until the number of blocks has been satisfied.
    """

    def __init__(
        self,
        in_channels,
        horizontal_block_factory,
        n_block=1,
        block_expansion_factor=2,
        expand_only_on_down=False,
        blocks_per_downsample=1,
        blocks_before_first_downsample=0,
        always_include_skip_connection=True,
        deepest_inner="identity",
        downsampling_modes="max",
        upsampling_modes="bilinear",
        _i_block=0,
        _i_down=0,
    ):
        super(UNetBlock, self).__init__()

        # Ensure these variables are a tuple of length two
        blocks_per_downsample = modules.utils._pair(blocks_per_downsample)
        blocks_before_first_downsample = modules.utils._pair(
            blocks_before_first_downsample
        )

        # Check which downsampling and upsampling mode we are using for this
        # layer (may be the same for every layer)
        if isinstance(downsampling_modes, str):
            downsampling_mode = downsampling_modes
        elif _i_down >= len(downsampling_modes):
            downsampling_mode = downsampling_modes[-1]
        else:
            downsampling_mode = downsampling_modes[_i_down]

        if isinstance(upsampling_modes, str):
            upsampling_mode = upsampling_modes
        elif _i_down >= len(upsampling_modes):
            upsampling_mode = upsampling_modes[-1]
        else:
            upsampling_mode = upsampling_modes[_i_down]

        # Check which dimensions need to be compressed with this block
        compress_dims = tuple(
            _i_block >= i0 and (_i_block - i0) % k == 0
            for i0, k in zip(blocks_before_first_downsample, blocks_per_downsample)
        )
        compress_any_dims = any(compress_dims)

        # Determine whether we are increasing the number of channels, and
        # if so what to
        if expand_only_on_down and not compress_any_dims:
            out_channels = in_channels
        else:
            out_channels = int(max(1, round(in_channels * block_expansion_factor)))

        # Downsamling step. If the mode is "stride", this is incorporated
        # into the horizontal block with a strided convolution. If there
        # is no need to downsample, it will be the Identity function.
        stride = 1
        if not compress_any_dims:
            self.down = nn.Identity()
        elif downsampling_mode == "stride":
            self.down = nn.Identity()
            stride = tuple(2 if compress_dim else 1 for compress_dim in compress_dims)
        else:
            self.down = Down(mode=downsampling_mode, compress_dims=compress_dims)

        # First horizontal block. It might begin with a downsampling stride,
        # and might increase the number of channels.
        self.horizontal_block_a = horizontal_block_factory(
            in_channels,
            out_channels,
            stride=stride,
        )

        # In the sequence, the inner step comes next. But we will define it
        # once we have finished defining everything else in this UNet block.

        # Upsampling step. Does the inverse of the Down step, using some
        # method.
        if not compress_any_dims:
            self.up = nn.Identity()
        else:
            self.up = Up(
                in_channels=out_channels, up_dims=compress_dims, mode=upsampling_mode
            )

        if compress_any_dims or always_include_skip_connection:
            # Concatenation step
            self.concatenate = modules.FlexibleConcat2d()
            b_in_channels = in_channels + out_channels
        else:
            # No concatenation step
            self.concatenate = None
            b_in_channels = out_channels

        # Second horizontal block. Takes both the skip connection and the
        # upsampled data as its input.
        self.horizontal_block_b = horizontal_block_factory(b_in_channels, in_channels)

        if _i_block + 1 < n_block:
            # Recurse deeper! Call this class again, but with the
            # block counter increased.
            self.nested = UNetBlock(
                out_channels,
                horizontal_block_factory,
                n_block=n_block,
                block_expansion_factor=block_expansion_factor,
                expand_only_on_down=expand_only_on_down,
                blocks_per_downsample=blocks_per_downsample,
                blocks_before_first_downsample=blocks_before_first_downsample,
                always_include_skip_connection=always_include_skip_connection,
                deepest_inner=deepest_inner,
                downsampling_modes=downsampling_modes,
                upsampling_modes=upsampling_modes,
                _i_block=_i_block + 1,
                _i_down=_i_down + compress_any_dims,
            )
        elif callable(deepest_inner):
            self.nested = deepest_inner
        elif deepest_inner is None or deepest_inner.lower() == "identity":
            # End recursion, by doing nothing for the inner loop.
            self.nested = nn.Identity()
        elif deepest_inner == "horizontal_block":
            # End recursion, by doing an extra regular block.
            self.nested = horizontal_block_factory(out_channels, out_channels)
        else:
            raise ValueError(
                "Unsupported deepest_inner value: {}".format(deepest_inner)
            )

    def forward(self, input):
        x = self.down(input)
        x = self.horizontal_block_a(x)
        x = self.nested(x)
        x = self.up(x)
        if self.concatenate is not None:
            x = self.concatenate(x, input)
        x = self.horizontal_block_b(x)
        return x


class UNet(nn.Module):
    """
    UNet model.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    initial_channels : int, optional
        Number of latent channels to output from the initial convolution
        facing the input layer. Default is ``32``.
    bottleneck_channels : int, optional
        Number of channels to output from the first block, before the first
        unet downsampling step can occur. Default is the same as
        ``initial_channels``.
    n_block : int, optional
        Number of blocks, both up and down. Default is ``4``.
    unet_expansion_factor : int or float, optional
        Channel expansion factor between unet blocks. Default is ``2``.
    expand_only_on_down : bool, optional
        Whether to only apply ``unet_expansion_factor`` on unet blocks which
        actually containg a down/up sampling component, and not on vanilla
        blocks. Default is ``False``.
    blocks_per_downsample : int or sequence, optional
        Block interval between dowsampling steps in the unet. If this is
        a sequence, it corresponds to the number of blocks for each spatial
        dimension. Default is ``1``.
    blocks_before_first_downsample : int, optional
        Number of blocks to use before and after the main unet structure.
        Must be at least ``1``. Default is ``1``.
    always_include_skip_connection : bool, optional
        If ``True``, a skip connection is included between all blocks equally
        far from the start and end of the UNet. If ``False``, skip connections
        are only used between downsampling and upsampling operations. Default
        is ``True``.
    deepest_inner : {callable, "horizontal_block", "identity", None}, optional
        A layer which should be applied at the deepest part of the network,
        before the first upsampling step. The parameter should either be a
        pre-instantiated layer, or the string ``"horizontal_block"``, to indicate
        an additional block as generated by the ``horizontal_block_factory``.
        If it is the string ``"identity"`` or ``None`` (default), no additional
        layer is included at the deepest point before upsampling begins.
    intrablock_expansion : int or float, optional
        Channel expansion factor within inverse residual block. Default is ``6``.
    se_reduction : int or float, optional
        Channel reduction factor within squeeze and excite block.
        Default is ``4``.
    downsampling_modes : {"max", "avg", "stride"} or sequence, optional
        The downsampling mode to use. If this is a string, the same
        downsampling mode is used for every downsampling step. If it is
        a sequence, it should contain a string for each downsampling step.
        If the input sequence is too short, the final value will be used
        for all remaining downsampling steps. Default is ``"max"``.
    upsampling_modes : str or sequence, optional
        The upsampling mode to use. If this is a string, it must be ``"conv"``,
        or something supported by :class:`torch.nn.Upsample`; the same
        upsampling mode is used for every upsampling step. If it is
        a sequence, it should contain a string for each upsampling step.
        If the input sequence is too short, the final value will be used
        for all remaining upsampling steps. Default is ``"bilinear"``.
    depthwise_separable_conv : bool, optional
        Whether to use depthwise separable convolutions in the MBConv block.
        Otherwise, the depth and pointwise convolutions are fused together
        into a regular convolution. Default is ``True``.
    residual : bool, optional
        Whether to use a residual architecture for the MBConv blocks.
        Default is ``True``.
    actfn : str, optional
        Name of the activation function to use. Default is ``"InplaceReLU"``.
    kernel_size : int, optional
        Size of convolution kernel to use. Default is ``5``.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        initial_channels=32,
        bottleneck_channels=None,
        n_block=4,
        unet_expansion_factor=2,
        expand_only_on_down=False,
        blocks_per_downsample=1,
        blocks_before_first_downsample=1,
        always_include_skip_connection=True,
        deepest_inner="identity",
        intrablock_expansion=6,
        se_reduction=4,
        downsampling_modes="max",
        upsampling_modes="bilinear",
        depthwise_separable_conv=True,
        residual=True,
        actfn="InplaceReLU",
        kernel_size=5,
    ):
        super(UNet, self).__init__()

        if bottleneck_channels is None:
            bottleneck_channels = initial_channels

        blocks_before_first_downsample = modules.utils._pair(
            blocks_before_first_downsample
        )

        if any(b < 1 for b in blocks_before_first_downsample):
            raise ValueError(
                "An initial block is hard coded. Number of blocks before first"
                " downsample must be at least 1."
            )

        self.in_channels = in_channels
        self.out_channels = out_channels

        actfn_factory = modules.activations.str2actfnfactory(actfn)

        horizontal_block_factory = functools.partial(
            modules.MBConv,
            expansion=intrablock_expansion,
            se_reduction=se_reduction,
            fused=not depthwise_separable_conv,
            residual=residual,
            actfn=actfn_factory,
            kernel_size=kernel_size,
        )

        self.initial_conv = nn.Sequential(
            modules.Conv2dSame(in_channels, initial_channels, kernel_size=kernel_size),
            nn.BatchNorm2d(initial_channels),
            actfn_factory(),
        )
        self.first_block = horizontal_block_factory(
            initial_channels,
            bottleneck_channels,
            expansion=1,
        )
        self.main_blocks = UNetBlock(
            bottleneck_channels,
            horizontal_block_factory,
            n_block=n_block,
            block_expansion_factor=unet_expansion_factor,
            expand_only_on_down=expand_only_on_down,
            blocks_per_downsample=blocks_per_downsample,
            blocks_before_first_downsample=tuple(
                b - 1 for b in blocks_before_first_downsample
            ),
            always_include_skip_connection=always_include_skip_connection,
            deepest_inner=deepest_inner,
            downsampling_modes=downsampling_modes,
            upsampling_modes=upsampling_modes,
        )
        self.final_block = horizontal_block_factory(bottleneck_channels, out_channels)

    def forward(self, x):
        x = self.initial_conv(x)
        x = self.first_block(x)
        x = self.main_blocks(x)
        x = self.final_block(x)
        return x



"""
Pytorch classes backported from later versions.

This contains functions copied from newer versions of pytorch than v1.2.0,
which is the latest version currently available from IBM compiled for ppc64
architectures.


From PyTorch:

Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
Copyright (c) 2011-2013 NYU                      (Clement Farabet)
Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)

From Caffe2:

Copyright (c) 2016-present, Facebook Inc. All rights reserved.

All contributions by Facebook:
Copyright (c) 2016 Facebook Inc.

All contributions by Google:
Copyright (c) 2015 Google Inc.
All rights reserved.

All contributions by Yangqing Jia:
Copyright (c) 2015 Yangqing Jia
All rights reserved.

All contributions from Caffe:
Copyright(c) 2013, 2014, 2015, the respective contributors
All rights reserved.

All other contributions:
Copyright(c) 2015, 2016 the respective contributors
All rights reserved.

Caffe2 uses a copyright model similar to Caffe: each contributor holds
copyright over their contributions to Caffe2. The project versioning records
all such contribution and copyright details. If a contributor wants to further
mark their specific copyright on a particular contribution, they should
indicate their copyright solely in the commit message of the change when it is
committed.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
   and IDIAP Research Institute nor the names of its contributors may be
   used to endorse or promote products derived from this software without
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
"""

import math
import warnings
import weakref
from collections import Counter
from functools import wraps

from torch._six import inf
from torch.optim.optimizer import Optimizer


class _LRScheduler(object):
    r"""
    Backported from pytorch 1.4.0.
    """

    def __init__(self, optimizer, last_epoch=-1):

        # Attach optimizer
        if not isinstance(optimizer, Optimizer):
            raise TypeError("{} is not an Optimizer".format(type(optimizer).__name__))
        self.optimizer = optimizer

        # Initialize epoch and base learning rates
        if last_epoch == -1:
            for group in optimizer.param_groups:
                group.setdefault("initial_lr", group["lr"])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if "initial_lr" not in group:
                    raise KeyError(
                        "param 'initial_lr' is not specified "
                        "in param_groups[{}] when resuming an optimizer".format(i)
                    )
        self.base_lrs = list(  # noqa: C417
            map(lambda group: group["initial_lr"], optimizer.param_groups)  # noqa: C417
        )  # noqa: C417
        self.last_epoch = last_epoch

        # Following https://github.com/pytorch/pytorch/issues/20124
        # We would like to ensure that `lr_scheduler.step()` is called after
        # `optimizer.step()`
        def with_counter(method):
            if getattr(method, "_with_counter", False):
                # `optimizer.step()` has already been replaced, return.
                return method

            # Keep a weak reference to the optimizer instance to prevent
            # cyclic references.
            instance_ref = weakref.ref(method.__self__)
            # Get the unbound method for the same purpose.
            func = method.__func__
            cls = instance_ref().__class__
            del method

            @wraps(func)
            def wrapper(*args, **kwargs):
                instance = instance_ref()
                instance._step_count += 1
                wrapped = func.__get__(instance, cls)
                return wrapped(*args, **kwargs)

            # Note that the returned function here is no longer a bound method,
            # so attributes like `__func__` and `__self__` no longer exist.
            wrapper._with_counter = True
            return wrapper

        self.optimizer.step = with_counter(self.optimizer.step)
        self.optimizer._step_count = 0
        self._step_count = 0

        self.step()

    def state_dict(self):
        """Return the state of the scheduler as a :class:`dict`.

        It contains an entry for every variable in self.__dict__ which
        is not the optimizer.
        """
        return {
            key: value for key, value in self.__dict__.items() if key != "optimizer"
        }

    def load_state_dict(self, state_dict):
        """Load the schedulers state.

        Arguments
        ---------
        state_dict : dict
            Scheduler state. Should be an object returned from a call to :meth:`state_dict`.
        """
        self.__dict__.update(state_dict)

    def get_last_lr(self):
        """Return last computed learning rate by current scheduler."""
        return self._last_lr

    def get_lr(self):
        # Compute learning rate using chainable form of the scheduler
        raise NotImplementedError

    def step(self, epoch=None):
        # Raise a warning if old pattern is detected
        # https://github.com/pytorch/pytorch/issues/20124
        if self._step_count == 1:
            if not hasattr(self.optimizer.step, "_with_counter"):
                warnings.warn(
                    "Seems like `optimizer.step()` has been overridden after learning rate scheduler "
                    "initialization. Please, make sure to call `optimizer.step()` before "
                    "`lr_scheduler.step()`. See more details at "
                    "https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate",
                    UserWarning,
                )

            # Just check if there were two first lr_scheduler.step() calls before optimizer.step()
            elif self.optimizer._step_count < 1:
                warnings.warn(
                    "Detected call of `lr_scheduler.step()` before `optimizer.step()`. "
                    "In PyTorch 1.1.0 and later, you should call them in the opposite order: "
                    "`optimizer.step()` before `lr_scheduler.step()`.  Failure to do this "
                    "will result in PyTorch skipping the first value of the learning rate schedule. "
                    "See more details at "
                    "https://pytorch.org/docs/stable/optim.html#how-to-adjust-learning-rate",
                    UserWarning,
                )
        self._step_count += 1

        class _enable_get_lr_call:
            def __init__(self, o):
                self.o = o

            def __enter__(self):
                self.o._get_lr_called_within_step = True
                return self

            def __exit__(self, type, value, traceback):
                self.o._get_lr_called_within_step = False
                return self

        with _enable_get_lr_call(self):
            if epoch is None:
                self.last_epoch += 1
                values = self.get_lr()
            else:
                warnings.warn("Something deprecated!", DeprecationWarning)
                self.last_epoch = epoch
                if hasattr(self, "_get_closed_form_lr"):
                    values = self._get_closed_form_lr()
                else:
                    values = self.get_lr()

        for param_group, lr in zip(self.optimizer.param_groups, values):
            param_group["lr"] = lr

        self._last_lr = [group["lr"] for group in self.optimizer.param_groups]


class OneCycleLR(_LRScheduler):
    r"""
    Backported from pytorch 1.4.0.

    Sets the learning rate of each parameter group according to the
    1cycle learning rate policy. The 1cycle policy anneals the learning
    rate from an initial learning rate to some maximum learning rate and then
    from that maximum learning rate to some minimum learning rate much lower
    than the initial learning rate.
    This policy was initially described in the paper `Super-Convergence:
    Very Fast Training of Neural Networks Using Large Learning Rates`_.

    The 1cycle learning rate policy changes the learning rate after every batch.
    ``step`` should be called after a batch has been used for training.

    This scheduler is not chainable.

    Note also that the total number of steps in the cycle can be determined in one
    of two ways (listed in order of precedence):

    #. A value for total_steps is explicitly provided.
    #. A number of epochs (epochs) and a number of steps per epoch
       (steps_per_epoch) are provided.
       In this case, the number of total steps is inferred by
       total_steps = epochs * steps_per_epoch

    You must either provide a value for total_steps or provide a value for both
    epochs and steps_per_epoch.

    Parameters
    ----------
    optimizer : :class:`torch.optim.optimizer.Optimizer`
        Wrapped optimizer.
    max_lr : float or list
        Upper learning rate boundaries in the cycle
        for each parameter group.
    total_steps : int
        The total number of steps in the cycle. Note that
        if a value is provided here, then it must be inferred by providing
        a value for epochs and steps_per_epoch.
        Default: None
    epochs : int
        The number of epochs to train for. This is used along
        with steps_per_epoch in order to infer the total number of steps in the cycle
        if a value for total_steps is not provided.
        Default: None
    steps_per_epoch : int
        The number of steps per epoch to train for. This is
        used along with epochs in order to infer the total number of steps in the
        cycle if a value for total_steps is not provided.
        Default: None
    pct_start : float
        The percentage of the cycle (in number of steps) spent
        increasing the learning rate.
        Default: 0.3
    anneal_strategy : {'cos', 'linear'}
        Specifies the annealing strategy: "cos" for cosine annealing, "linear" for
        linear annealing.
        Default: 'cos'
    cycle_momentum : bool
        If ``True``, momentum is cycled inversely
        to learning rate between 'base_momentum' and 'max_momentum'.
        Default: True
    base_momentum : float or list
        Lower momentum boundaries in the cycle
        for each parameter group. Note that momentum is cycled inversely
        to learning rate; at the peak of a cycle, momentum is
        'base_momentum' and learning rate is 'max_lr'.
        Default: 0.85
    max_momentum : float or list
        Upper momentum boundaries in the cycle
        for each parameter group. Functionally,
        it defines the cycle amplitude (max_momentum - base_momentum).
        Note that momentum is cycled inversely
        to learning rate; at the start of a cycle, momentum is 'max_momentum'
        and learning rate is 'base_lr'
        Default: 0.95
    div_factor : float
        Determines the initial learning rate via
        initial_lr = max_lr/div_factor
        Default: 25
    final_div_factor : float
        Determines the minimum learning rate via
        min_lr = initial_lr/final_div_factor
        Default: 1e4
    last_epoch : int
        The index of the last batch. This parameter is used when
        resuming a training job. Since ``step()`` should be invoked after each
        batch instead of after each epoch, this number represents the total
        number of *batches* computed, not the total number of epochs computed.
        When last_epoch=-1, the schedule is started from the beginning.
        Default: -1

    Example
    -------
    >>> data_loader = torch.utils.data.DataLoader(...)
    >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.1, momentum=0.9)
    >>> scheduler = torch.optim.lr_scheduler.OneCycleLR(
    >>>     optimizer, max_lr=0.01, steps_per_epoch=len(data_loader), epochs=10
    >>> )
    >>> for epoch in range(10):
    >>>     for batch in data_loader:
    >>>         train_batch(...)
    >>>         scheduler.step()

    .. _Super-Convergence\: Very Fast Training of Neural Networks Using Large Learning Rates:
        https://arxiv.org/abs/1708.07120
    """

    def __init__(
        self,
        optimizer,
        max_lr,
        total_steps=None,
        epochs=None,
        steps_per_epoch=None,
        pct_start=0.3,
        anneal_strategy="cos",
        cycle_momentum=True,
        base_momentum=0.85,
        max_momentum=0.95,
        div_factor=25.0,
        final_div_factor=1e4,
        last_epoch=-1,
    ):

        # Validate optimizer
        if not isinstance(optimizer, Optimizer):
            raise TypeError("{} is not an Optimizer".format(type(optimizer).__name__))
        self.optimizer = optimizer

        # Validate total_steps
        if total_steps is None and epochs is None and steps_per_epoch is None:
            raise ValueError(
                "You must define either total_steps OR (epochs AND steps_per_epoch)"
            )
        elif total_steps is not None:
            if total_steps <= 0 or not isinstance(total_steps, int):
                raise ValueError(
                    "Expected non-negative integer total_steps, but got {}".format(
                        total_steps
                    )
                )
            self.total_steps = total_steps
        else:
            if epochs <= 0 or not isinstance(epochs, int):
                raise ValueError(
                    "Expected non-negative integer epochs, but got {}".format(epochs)
                )
            if steps_per_epoch <= 0 or not isinstance(steps_per_epoch, int):
                raise ValueError(
                    "Expected non-negative integer steps_per_epoch, but got {}".format(
                        steps_per_epoch
                    )
                )
            self.total_steps = epochs * steps_per_epoch
        self.step_size_up = float(pct_start * self.total_steps) - 1
        self.step_size_down = float(self.total_steps - self.step_size_up) - 1

        # Validate pct_start
        if pct_start < 0 or pct_start > 1 or not isinstance(pct_start, float):
            raise ValueError(
                "Expected float between 0 and 1 pct_start, but got {}".format(pct_start)
            )

        # Validate anneal_strategy
        if anneal_strategy not in ["cos", "linear"]:
            raise ValueError(
                "anneal_strategy must by one of 'cos' or 'linear', instead got {}".format(
                    anneal_strategy
                )
            )
        elif anneal_strategy == "cos":
            self.anneal_func = self._annealing_cos
        elif anneal_strategy == "linear":
            self.anneal_func = self._annealing_linear

        # Initialize learning rate variables
        max_lrs = self._format_param("max_lr", self.optimizer, max_lr)
        if last_epoch == -1:
            for idx, group in enumerate(self.optimizer.param_groups):
                group["initial_lr"] = max_lrs[idx] / div_factor
                group["max_lr"] = max_lrs[idx]
                group["min_lr"] = group["initial_lr"] / final_div_factor

        # Initialize momentum variables
        self.cycle_momentum = cycle_momentum
        if self.cycle_momentum:
            if (
                "momentum" not in self.optimizer.defaults
                and "betas" not in self.optimizer.defaults
            ):
                raise ValueError(
                    "optimizer must support momentum with `cycle_momentum` option enabled"
                )
            self.use_beta1 = "betas" in self.optimizer.defaults
            max_momentums = self._format_param("max_momentum", optimizer, max_momentum)
            base_momentums = self._format_param(
                "base_momentum", optimizer, base_momentum
            )
            if last_epoch == -1:
                for m_momentum, b_momentum, group in zip(
                    max_momentums, base_momentums, optimizer.param_groups
                ):
                    if self.use_beta1:
                        _, beta2 = group["betas"]
                        group["betas"] = (m_momentum, beta2)
                    else:
                        group["momentum"] = m_momentum
                    group["max_momentum"] = m_momentum
                    group["base_momentum"] = b_momentum

        super(OneCycleLR, self).__init__(optimizer, last_epoch)

    def _format_param(self, name, optimizer, param):
        """Return correctly formatted lr/momentum for each param group."""
        if isinstance(param, (list, tuple)):
            if len(param) != len(optimizer.param_groups):
                raise ValueError(
                    "expected {} values for {}, got {}".format(
                        len(optimizer.param_groups), name, len(param)
                    )
                )
            return param
        else:
            return [param] * len(optimizer.param_groups)

    def _annealing_cos(self, start, end, pct):
        """Cosine anneal from ``start`` to ``end`` as pct goes from 0.0 to 1.0."""
        cos_out = math.cos(math.pi * pct) + 1
        return end + (start - end) / 2.0 * cos_out

    def _annealing_linear(self, start, end, pct):
        """Linearly anneal from ``start`` to ``end`` as pct goes from 0.0 to 1.0."""
        return (end - start) * pct + start

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, "
                "please use `get_last_lr()`.",
                DeprecationWarning,
            )

        lrs = []
        step_num = self.last_epoch

        if step_num > self.total_steps:
            raise ValueError(
                "Tried to step {} times. The specified number of total steps is {}".format(
                    step_num + 1, self.total_steps
                )
            )

        for group in self.optimizer.param_groups:
            if step_num <= self.step_size_up:
                computed_lr = self.anneal_func(
                    group["initial_lr"], group["max_lr"], step_num / self.step_size_up
                )
                if self.cycle_momentum:
                    computed_momentum = self.anneal_func(
                        group["max_momentum"],
                        group["base_momentum"],
                        step_num / self.step_size_up,
                    )
            else:
                down_step_num = step_num - self.step_size_up
                computed_lr = self.anneal_func(
                    group["max_lr"],
                    group["min_lr"],
                    down_step_num / self.step_size_down,
                )
                if self.cycle_momentum:
                    computed_momentum = self.anneal_func(
                        group["base_momentum"],
                        group["max_momentum"],
                        down_step_num / self.step_size_down,
                    )

            lrs.append(computed_lr)
            if self.cycle_momentum:
                if self.use_beta1:
                    _, beta2 = group["betas"]
                    group["betas"] = (computed_momentum, beta2)
                else:
                    group["momentum"] = computed_momentum

        return lrs



"""
Loader utility functions.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import scipy.interpolate

from ..utils import get_indicator_onoffsets, mode


def interp1d_preserve_nan(
    x, y, x_samples, nan_threshold=0.0, bounds_error=False, **kwargs
):
    """
    Interpolate a 1-D function, preserving NaNs.

    Inputs ``x`` and ``y`` are arrays of values used to approximate some function f:
    ``y = f(x)``.  We exclude NaNs for the interpolation and then mask out
    entries which are adjacent (or close to) a NaN in the input.

    Parameters
    ----------
    x : (N,) array_like
        A 1-D array of real values. Must not contain NaNs.
    y : (...,N,...) array_like
        A N-D array of real values. The length of ``y`` along the interpolation
        axis must be equal to the length of ``x``. May contain NaNs.
    x_samples : array_like
        A 1-D array of real values at which the interpolation function will
        be sampled.
    nan_threshold : float, optional
        Minimum amount of influence a NaN must have on an output sample for it
        to become a NaN. Default is ``0.`` i.e. any influence.
    bounds_error : bool, optional
        If ``True``, a ValueError is raised any time interpolation is attempted
        on a value outside of the range of ``x`` (where extrapolation is
        necessary). If ``False`` (default), out of bounds values are assigned
        value ``fill_value`` (whose default is NaN).
    **kwargs
        Additional keyword arguments are as per :meth:`scipy.interpolate.interp1d`.

    Returns
    -------
    y_samples : (...,N,...) np.ndarray
        The result of interpolating, with sample points close to NaNs in the
        input returned as NaN.
    """
    # First, run with NaNs masked out.
    is_nan = np.isnan(y)
    if np.sum(~is_nan) < 2 and np.ndim(y) == 1:
        y_samples = np.empty(x_samples.shape, dtype=y.dtype)
        y_samples[:] = np.nan
        return y_samples
    y_samples = scipy.interpolate.interp1d(
        x[~is_nan], y[~is_nan], bounds_error=bounds_error, **kwargs
    )(x_samples)
    if np.sum(is_nan) == 0:
        # Shortcut if there are no NaNs
        return y_samples
    # Then find the points close to NaNs
    influence = scipy.interpolate.interp1d(
        x, is_nan, bounds_error=bounds_error, **kwargs
    )(x_samples)
    # and remove the points too close to a NaN in the input
    y_samples[influence > nan_threshold] = np.nan
    return y_samples


def pad1d(array, pad_width, axis=0, **kwargs):
    """
    Pad an array along a single axis only.

    Parameters
    ----------
    array : numpy.ndarary
        Array to be padded.
    pad_width : int or tuple
        The amount to pad, either a length two tuple of values for each edge,
        or an int if the padding should be the same for each side.
    axis : int, optional
        The axis to pad. Default is ``0``.
    **kwargs
        As per :meth:`numpy.pad`.

    Returns
    -------
    numpy.ndarary
        Padded array.

    See Also
    --------
    numpy.pad
    """
    pads = [(0, 0) for _ in range(array.ndim)]
    if hasattr(pad_width, "__len__"):
        pads[axis] = pad_width
    else:
        pads[axis] = (pad_width, pad_width)
    return np.pad(array, pads, **kwargs)


def medfilt1d(signal, kernel_size, axis=-1, pad_mode="reflect"):
    """
    Median filter in 1d, with support for selecting padding mode.

    Parameters
    ----------
    signal : array_like
        The signal to filter.
    kernel_size
        Size of the median kernel to use.
    axis : int, optional
        Which axis to operate along. Default is ``-1``.
    pad_mode : str, optional
        Method with which to pad the vector at the edges.
        Must be supported by :meth:`numpy.pad`. Default is ``"reflect"``.

    Returns
    -------
    filtered : array_like
        The filtered signal.

    See Also
    --------
    scipy.signal.medfilt
    pad1d
    """
    offset = kernel_size // 2
    signal = pad1d(signal, offset, axis=axis, mode=pad_mode)
    filtered = scipy.signal.medfilt(signal, kernel_size)[offset:-offset]
    return filtered


def squash_gaps(mask, max_gap_squash, axis=-1, inplace=False):
    """
    Merge small gaps between zero values in a boolean array.

    Parameters
    ----------
    mask : boolean array
        The input mask, with small gaps between zero values which will be
        squashed with zeros.
    max_gap_squash : int
        Maximum length of gap to squash.
    axis : int, optional
        Axis on which to operate. Default is ``-1``.
    inplace : bool, optional
        Whether to operate on the original array. If ``False``, a copy is
        created and returned.

    Returns
    -------
    merged_mask : boolean array
        Mask as per the input, but with small gaps squashed.
    """
    if not inplace:
        mask = mask.copy()
    L = mask.shape[axis]
    for i in range(min(max_gap_squash, L - 1), 1, -1):
        check = np.stack(
            [
                pad1d(
                    mask.take(range(i // 2, L), axis=axis),
                    (0, i // 2),
                    axis=axis,
                    mode="constant",
                ),
                pad1d(
                    mask.take(range(0, L - ((i + 1) // 2)), axis=axis),
                    ((i + 1) // 2, 0),
                    axis=axis,
                    mode="constant",
                ),
            ]
        )
        li = ~np.any(check, axis=0)
        mask[li] = 0
    return mask


def integrate_area_of_contour(x, y, closed=None, preserve_sign=False):
    """
    Compute the area within a contour, using Green's algorithm.

    Parameters
    ----------
    x : array_like vector
        x co-ordinates of nodes along the contour.
    y : array_like vector
        y co-ordinates of nodes along the contour.
    closed : bool or None, optional
        Whether the contour is already closed. If ``False``, it will be closed
        before deterimining the area. If ``None`` (default), it is automatically
        determined as to whether the contour is already closed, and is closed
        if necessary.
    preserve_sign : bool, optional
        Whether to preserve the sign of the area. If ``True``, the area is
        positive if the contour is anti-clockwise and negative if it is
        clockwise oriented. Default is ``False``, which always returns a positive
        area.

    Returns
    -------
    area : float
        The integral of the area witihn the contour.

    Notes
    -----
    https://en.wikipedia.org/wiki/Green%27s_theorem#Area_calculation
    """
    if closed is None:
        closed = x[0] == x[-1] and y[0] == y[-1]
    if not closed:
        x = np.concatenate([x, x[[0]]])
        y = np.concatenate([y, y[[0]]])
    # Integrate to find the area
    A = 0.5 * np.sum(y[:-1] * np.diff(x) - x[:-1] * np.diff(y))
    # Take the abs in case the curve was clockwise instead of anti-clockwise
    A = np.abs(A)
    return A


def fillholes2d(arr, nan_thr=2, interp_method="linear", inplace=False):
    """
    Interpolate to replace NaN values in 2d gridded array data.

    Parameters
    ----------
    arr : 2d numpy.ndarray
        Array in 2d which, may contain NaNs.
    nan_thr : int, default=2
        Minimum number of NaN values needed in a row/column for it
        to be included in the (rectangular) area where NaNs are fixed.
    interp_method : str, default="linear"
        Interpolation method.
    inplace : bool, default=False
        Whether to update arr instead of a copy.

    Returns
    -------
    arr : 2d numpy.ndarray
        Like input ``arr``, but with NaN values replaced with
        interpolated values.
    """
    # Find where the nans are
    isna = np.isnan(arr[1:-1])
    x_extent = np.nonzero(isna.sum(1) >= nan_thr)[0]
    y_extent = np.nonzero(isna.sum(0) >= nan_thr)[0]
    if len(x_extent) < 1 and len(y_extent) < 1:
        # Nothing to do
        return arr
    if len(x_extent) < 1:
        # There is y-axis things to do, so be more lenient with x
        x_extent = np.nonzero(isna.any(1))[0]
    if len(y_extent) < 1:
        # There is x-axis things to do, so be more lenient with y
        y_extent = np.nonzero(isna.any(0))[0]
    if len(x_extent) < 1 or len(y_extent) < 1:
        # Nothing to do
        return arr
    x_extent = x_extent[[0, -1]] + 1
    y_extent = y_extent[[0, -1]] + 1

    # We will extract an area bigger than the space covering all NaNs
    x_extent += [-4, 4]
    y_extent += [-4, 4]
    x_extent = np.minimum(arr.shape[0] - 1, np.maximum(0, x_extent))
    y_extent = np.minimum(arr.shape[1] - 1, np.maximum(0, y_extent))

    # Hanndle indexing the array
    z_source = arr[x_extent[0] : x_extent[1] + 1, y_extent[0] : y_extent[1] + 1]
    x_source = np.arange(x_extent[0], x_extent[1] + 1)
    y_source = np.arange(y_extent[0], y_extent[1] + 1)

    xx, yy = np.meshgrid(x_source, y_source)
    xx = xx.T
    yy = yy.T

    # Find all the NaNs. This tells us where to update and the inverse tells us
    # where to take training values for the interpolation
    is_missing = np.isnan(z_source)
    in_coords = np.stack([xx[~is_missing], yy[~is_missing]], -1)
    out_coords = np.stack([xx[is_missing], yy[is_missing]], -1)

    # Perform 2d interpolation
    z_out = scipy.interpolate.griddata(
        in_coords,
        z_source[~is_missing],
        out_coords,
        method=interp_method,
    )
    # Convert 2d coordinates into linear indices for put
    out_indices = np.ravel_multi_index((out_coords[:, 0], out_coords[:, 1]), arr.shape)

    # Insert the interpolated values at their locations
    if not inplace:
        arr = arr.copy()
    np.put(arr, out_indices, z_out)

    return arr



"""
Provides extensions to argparse.
"""

# This file is part of Echofilter.
#
# Copyright (C) 2020-2022  Scott C. Lowe and Offshore Energy Research Association (OERA)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse


class DedentTextHelpFormatter(argparse.HelpFormatter):
    """
    Help message formatter.

    Retains formatting of all help text, except from indentation.
    Leading new lines are also stripped.
    """

    def _split_lines(self, text, width):
        import textwrap

        return textwrap.dedent(text.lstrip("\n")).splitlines()

    def _fill_text(self, text, width, indent):
        import textwrap

        return "".join(
            indent + line
            for line in textwrap.dedent(text.lstrip("\n")).splitlines(keepends=True)
        )


class FlexibleHelpFormatter(argparse.HelpFormatter):
    """
    Help message formatter which can handle different formatting specifications.

    The following formatters are supported:

    ``"R|"``
        Raw. will be left as is, processed using
        ``argparse.RawTextHelpFormatter``.
    ``"d|"``
        Raw except for indentation. Will be dedented and leading
        newlines stripped only, processed using
        ``argparse.RawTextHelpFormatter``.

    The format specifier will be stripped from the text.

    Notes
    -----
    Based on https://stackoverflow.com/a/22157266/1960959
    and https://sourceforge.net/projects/ruamel-std-argparse/.
    """

    def _split_lines(self, text, *args, **kwargs):
        if len(text) < 2 or text[1] != "|":
            return super()._split_lines(text, *args, **kwargs)
        if text[0] == "R":
            return argparse.RawTextHelpFormatter._split_lines(
                self, text[2:], *args, **kwargs
            )
        if text[0] == "d":
            return DedentTextHelpFormatter._split_lines(self, text[2:], *args, **kwargs)
        raise ValueError("Invalid format code: {}".format(text[0]))

    def _fill_text(self, text, *args, **kwargs):
        if len(text) < 2 or text[1] != "|":
            return super()._fill_text(text, *args, **kwargs)
        if text[0] == "R":
            return argparse.RawTextHelpFormatter._fill_text(
                self, text[2:], *args, **kwargs
            )
        if text[0] == "d":
            return DedentTextHelpFormatter._fill_text(self, text[2:], *args, **kwargs)
        raise ValueError("Invalid format code: {}".format(text[0]))


def format_parser_for_sphinx(parser):
    """
    Pre-format parser help for sphinx-argparse processing.

    Parameters
    ----------
    parser : argparse.ArgumentParser
        Initial argument parser.

    Returns
    -------
    parser : argparse.ArgumentParser
        The same argument parser, but with raw help text touched up so it
        renders correctly when passed through sphinx-argparse.
    """
    for action_group in parser._action_groups:
        for action in action_group._group_actions:
            # Get the help text for this action
            help = action.help
            # Remove quotes around default strings, to prevent double-marking.
            help = help.replace('"%(default)s"', "%(default)s")
            help = help.replace("'%(default)s'", "%(default)s")
            # Wrap any default values in `` so they are rendered as code.
            help = help.replace("%(default)s", "``%(default)s``")
            # But also ensure we don't add backticks around backticks
            help = help.replace("````", "``")
            # Remove flexible formatter indictor, if present
            if (
                parser.formatter_class == FlexibleHelpFormatter
                and len(help) > 1
                and help[1] == "|"
            ):
                help = help[2:]
            # Overwrite the help text
            action.help = help
    return parser
